class UnitIdentityException(Exception):

    def __init__(self, message):
        self.message = message


class ScaleIdentityException(Exception):

    def __init__(self, message):
        self.message = message



import unittest

import requests_mock

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import Database, DatabaseAccess, AccessType, DatabaseBrief, UserBrief, \
    ContainerBrief, ImageBrief
from dbrepo.api.exceptions import ResponseCodeError, NotExistsError, ForbiddenError, MalformedError, \
    AuthenticationError, QueryStoreError, ServiceConnectionError, ServiceError


class DatabaseUnitTest(unittest.TestCase):

    def test_get_databases_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database', json=[])
            # test
            response = RestClient().get_databases()
            self.assertEqual([], response)

    def test_get_databases_succeeds(self):
        exp = [
            DatabaseBrief(
                id="6bd39359-b154-456d-b9c2-caa516a45732",
                name='test',
                owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16',
                contact=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                internal_name='test_abcd',
                is_public=True,
                is_schema_public=True
            )
        ]
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_databases()
            self.assertEqual(exp, response)

    def test_get_databases_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database', status_code=401)
            # test
            try:
                RestClient().get_databases()
            except ResponseCodeError:
                pass

    def test_get_databases_count_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head('/api/v1/database', headers={'X-Count': '100'})
            # test
            response = RestClient().get_databases_count()
            self.assertEqual(100, response)

    def test_get_databases_count_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head('/api/v1/database', status_code=401)
            # test
            try:
                RestClient().get_databases_count()
            except ResponseCodeError:
                pass

    def test_get_database_succeeds(self):
        exp = Database(
            id="6bd39359-b154-456d-b9c2-caa516a45732",
            name='test',
            owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
            contact=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
            exchange_name='dbrepo',
            internal_name='test_abcd',
            is_public=True,
            is_schema_public=True,
            is_dashboard_enabled=True,
            container=ContainerBrief(
                id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                name='MariaDB Galera 11.1.3',
                internal_name='mariadb',
                image=ImageBrief(
                    id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                    name='mariadb',
                    version='11.2.2',
                    default=True)
            )
        )
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732', json=exp.model_dump())
            # test
            response = RestClient().get_database(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            self.assertEqual(exp, response)

    def test_get_database_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732', status_code=403)
            # test
            try:
                RestClient().get_database(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ForbiddenError as e:
                pass

    def test_get_database_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732', status_code=404)
            # test
            try:
                RestClient().get_database(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except NotExistsError as e:
                pass

    def test_get_database_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732', status_code=202)
            # test
            try:
                RestClient().get_database(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ResponseCodeError as e:
                pass

    def test_create_database_succeeds(self):
        exp = Database(
            id="6bd39359-b154-456d-b9c2-caa516a45732",
            name='test',
            owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
            contact=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
            exchange_name='dbrepo',
            internal_name='test_abcd',
            is_public=True,
            is_schema_public=True,
            is_dashboard_enabled=True,
            container=ContainerBrief(
                id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                name='MariaDB Galera 11.1.3',
                internal_name='mariadb',
                image=ImageBrief(
                    id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                    name='mariadb',
                    version='11.2.2',
                    default=True)
            )
        )
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database', json=exp.model_dump(), status_code=201)
            # test
            client = RestClient(username="a", password="b")
            response = RestClient(username="a", password="b").create_database(name='test',
                                                                              container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                                                                              is_public=True)
            self.assertEqual(response.name, 'test')

    def test_create_database_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database', status_code=400)
            # test
            try:
                RestClient(username="a", password="b").create_database(name='test',
                                                                       container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                                                                       is_public=True)
            except MalformedError as e:
                pass

    def test_create_database_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database', status_code=403)
            # test
            try:
                RestClient(username="a", password="b").create_database(name='test',
                                                                       container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                                                                       is_public=True)
            except ForbiddenError as e:
                pass

    def test_create_database_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database', status_code=404)
            # test
            try:
                RestClient(username="a", password="b").create_database(name='test',
                                                                       container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                                                                       is_public=True)
            except NotExistsError as e:
                pass

    def test_create_database_409_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database', status_code=409)
            # test
            try:
                RestClient(username="a", password="b").create_database(name='test',
                                                                       container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                                                                       is_public=True)
            except QueryStoreError as e:
                pass

    def test_create_database_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database', status_code=502)
            # test
            try:
                RestClient(username="a", password="b").create_database(name='test',
                                                                       container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                                                                       is_public=True)
            except ServiceConnectionError as e:
                pass

    def test_create_database_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database', status_code=503)
            # test
            try:
                RestClient(username="a", password="b").create_database(name='test',
                                                                       container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                                                                       is_public=True)
            except ServiceError as e:
                pass

    def test_create_database_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database', status_code=202)
            # test
            try:
                RestClient(username="a", password="b").create_database(name='test',
                                                                       container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                                                                       is_public=True)
            except ResponseCodeError as e:
                pass

    def test_create_database_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database', status_code=404)
            # test
            try:
                RestClient().create_database(name='test',
                                             container_id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                                             is_public=True)
            except AuthenticationError as e:
                pass

    def test_update_database_visibility_succeeds(self):
        exp = Database(
            id="6bd39359-b154-456d-b9c2-caa516a45732",
            name='test',
            owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
            contact=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
            exchange_name='dbrepo',
            internal_name='test_abcd',
            is_public=True,
            is_schema_public=True,
            is_dashboard_enabled=True,
            container=ContainerBrief(
                id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                name='MariaDB Galera 11.1.3',
                internal_name='mariadb',
                image=ImageBrief(
                    id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                    name='mariadb',
                    version='11.2.2',
                    default=True)
            )
        )
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/visibility', json=exp.model_dump(),
                     status_code=202)
            # test
            response = RestClient(username="a", password="b").update_database_visibility(
                database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                is_public=True,
                is_schema_public=True,
                is_dashboard_enabled=True)
            self.assertEqual(response.is_public, True)

    def test_update_database_visibility_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/visibility', status_code=400)
            # test
            try:
                RestClient(username="a", password="b").update_database_visibility(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    is_public=True,
                    is_schema_public=True,
                    is_dashboard_enabled=True)
            except MalformedError:
                pass

    def test_update_database_visibility_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/visibility', status_code=403)
            # test
            try:
                response = RestClient(username="a", password="b").update_database_visibility(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    is_public=True,
                    is_schema_public=True,
                    is_dashboard_enabled=True)
            except ForbiddenError:
                pass

    def test_update_database_visibility_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/visibility', status_code=404)
            # test
            try:
                response = RestClient(username="a", password="b").update_database_visibility(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    is_public=True,
                    is_schema_public=True,
                    is_dashboard_enabled=True)
            except NotExistsError:
                pass

    def test_update_database_visibility_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/visibility', status_code=502)
            # test
            try:
                RestClient(username="a", password="b").update_database_visibility(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    is_public=True,
                    is_schema_public=True,
                    is_dashboard_enabled=True)
            except ServiceConnectionError:
                pass

    def test_update_database_visibility_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/visibility', status_code=503)
            # test
            try:
                RestClient(username="a", password="b").update_database_visibility(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    is_public=True,
                    is_schema_public=True,
                    is_dashboard_enabled=True)
            except ServiceError:
                pass

    def test_update_database_visibility_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/visibility', status_code=200)
            # test
            try:
                RestClient(username="a", password="b").update_database_visibility(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    is_public=True,
                    is_schema_public=True,
                    is_dashboard_enabled=True)
            except ResponseCodeError:
                pass

    def test_update_database_visibility_anonymous_fails(self):
        # test
        try:
            RestClient().update_database_visibility(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                    is_public=True,
                                                    is_schema_public=True,
                                                    is_dashboard_enabled=True)
        except AuthenticationError:
            pass

    def test_update_database_owner_succeeds(self):
        exp = Database(
            id="6bd39359-b154-456d-b9c2-caa516a45732",
            name='test',
            owner=UserBrief(id='abdbf897-e599-4e5a-a3f0-7529884ea011', username='mweise'),
            contact=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
            exchange_name='dbrepo',
            internal_name='test_abcd',
            is_public=True,
            is_schema_public=True,
            is_dashboard_enabled=True,
            container=ContainerBrief(
                id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                name='MariaDB Galera 11.1.3',
                internal_name='mariadb',
                image=ImageBrief(
                    id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                    name='mariadb',
                    version='11.2.2',
                    default=True)
            )
        )
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/owner', json=exp.model_dump(), status_code=202)
            # test
            client = RestClient(username="a", password="b")
            response = RestClient(username="a", password="b").update_database_owner(
                database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                username='bar')
            self.assertEqual(response.owner.id, 'abdbf897-e599-4e5a-a3f0-7529884ea011')

    def test_update_database_owner_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/owner', status_code=400)
            # test
            try:
                RestClient(username="a", password="b").update_database_owner(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    username='bar')
            except MalformedError:
                pass

    def test_update_database_owner_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/owner', status_code=403)
            # test
            try:
                RestClient(username="a", password="b").update_database_owner(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    username='bar')
            except ForbiddenError:
                pass

    def test_update_database_owner_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/owner', status_code=404)
            # test
            try:
                RestClient(username="a", password="b").update_database_owner(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    username='bar')
            except NotExistsError:
                pass

    def test_update_database_owner_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/owner', status_code=502)
            # test
            try:
                RestClient(username="a", password="b").update_database_owner(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    username='bar')
            except ServiceConnectionError:
                pass

    def test_update_database_owner_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/owner', status_code=503)
            # test
            try:
                RestClient(username="a", password="b").update_database_owner(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    username='bar')
            except ServiceError:
                pass

    def test_update_database_owner_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/owner', status_code=200)
            # test
            try:
                RestClient(username="a", password="b").update_database_owner(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    username='bar')
            except ResponseCodeError:
                pass

    def test_update_database_owner_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/owner', status_code=404)
            # test
            try:
                RestClient().update_database_owner(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                   username='bar')
            except AuthenticationError:
                pass

    def test_update_database_schema_succeeds(self):
        exp = DatabaseBrief(
            id="6bd39359-b154-456d-b9c2-caa516a45732",
            name='test',
            owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16',
            contact=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
            internal_name='test_abcd',
            is_public=True,
            is_schema_public=True
        )
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', json=exp.model_dump())
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/view', json=exp.model_dump())
            # test
            response = RestClient(username='foo', password='bar').update_database_schema(
                database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            self.assertEqual(exp, response)

    def test_update_database_schema_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', status_code=400)
            # test
            try:
                RestClient(username='foo', password='bar').update_database_schema(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except MalformedError:
                pass

    def test_update_database_schema_view_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', json={}, status_code=200)
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/view', status_code=400)
            # test
            try:
                RestClient(username='foo', password='bar').update_database_schema(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except MalformedError:
                pass

    def test_update_database_schema_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', status_code=403)
            # test
            try:
                RestClient(username='foo', password='bar').update_database_schema(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ForbiddenError:
                pass

    def test_update_database_schema_view_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', json={}, status_code=200)
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/view', status_code=403)
            # test
            try:
                RestClient(username='foo', password='bar').update_database_schema(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ForbiddenError:
                pass

    def test_update_database_schema_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', status_code=404)
            # test
            try:
                RestClient(username='foo', password='bar').update_database_schema(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except NotExistsError:
                pass

    def test_update_database_schema_view_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', json={}, status_code=200)
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/view', status_code=404)
            # test
            try:
                RestClient(username='foo', password='bar').update_database_schema(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except NotExistsError:
                pass

    def test_update_database_schema_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', status_code=502)
            # test
            try:
                RestClient(username='foo', password='bar').update_database_schema(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ServiceConnectionError:
                pass

    def test_update_database_schema_view_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', json={}, status_code=200)
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/view', status_code=502)
            # test
            try:
                RestClient(username='foo', password='bar').update_database_schema(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ServiceConnectionError:
                pass

    def test_update_database_schema_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', status_code=503)
            # test
            try:
                RestClient(username='foo', password='bar').update_database_schema(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ServiceError:
                pass

    def test_update_database_schema_view_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', json={}, status_code=200)
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/view', status_code=503)
            # test
            try:
                RestClient(username='foo', password='bar').update_database_schema(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ServiceError:
                pass

    def test_update_database_schema_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', status_code=202)
            # test
            try:
                RestClient(username='foo', password='bar').update_database_schema(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ResponseCodeError:
                pass

    def test_update_database_schema_view_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/table', json={}, status_code=200)
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/metadata/view', status_code=202)
            # test
            try:
                RestClient(username='foo', password='bar').update_database_schema(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ResponseCodeError:
                pass

    def test_update_database_schema_anonymous_fails(self):
        # test
        try:
            RestClient().update_database_schema(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
        except AuthenticationError:
            pass

    def test_get_database_access_succeeds(self):
        exp = DatabaseAccess(type=AccessType.READ,
                             user=UserBrief(id='abdbf897-e599-4e5a-a3f0-7529884ea011', username='other'))
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access', json=exp.model_dump())
            # test
            response = RestClient().get_database_access(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            self.assertEqual(response, AccessType.READ)

    def test_get_database_access_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access', status_code=403)
            # test
            try:
                RestClient().get_database_access(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ForbiddenError:
                pass

    def test_get_database_access_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access', status_code=404)
            # test
            try:
                RestClient().get_database_access(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except NotExistsError:
                pass

    def test_get_database_access_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access', status_code=202)
            # test
            try:
                RestClient().get_database_access(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ResponseCodeError:
                pass

    def test_create_database_access_succeeds(self):
        exp = DatabaseAccess(type=AccessType.READ,
                             user=UserBrief(id='abdbf897-e599-4e5a-a3f0-7529884ea011', username='other'))
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                      json=exp.model_dump(),
                      status_code=202)
            # test
            response = RestClient(username="a", password="b").create_database_access(
                database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                type=AccessType.READ,
                username='bar')
            self.assertEqual(response, exp.type)

    def test_create_database_access_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                      status_code=400)
            # test
            try:
                RestClient(username="a", password="b").create_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    type=AccessType.READ,
                    username='bar')
            except MalformedError:
                pass

    def test_create_database_access_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                      status_code=400)
            # test
            try:
                RestClient().create_database_access(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                    type=AccessType.READ,
                                                    username='bar')
            except AuthenticationError:
                pass

    def test_create_database_access_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                      status_code=403)
            # test
            try:
                RestClient(username="a", password="b").create_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    type=AccessType.READ,
                    username='bar')
            except ForbiddenError:
                pass

    def test_create_database_access_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                      status_code=404)
            # test
            try:
                RestClient(username="a", password="b").create_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    type=AccessType.READ,
                    username='bar')
            except NotExistsError:
                pass

    def test_create_database_access_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                      status_code=502)
            # test
            try:
                RestClient(username="a", password="b").create_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    type=AccessType.READ,
                    username='bar')
            except ServiceConnectionError:
                pass

    def test_create_database_access_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                      status_code=503)
            # test
            try:
                RestClient(username="a", password="b").create_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    type=AccessType.READ,
                    username='bar')
            except ServiceError:
                pass

    def test_create_database_access_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                      status_code=200)
            # test
            try:
                RestClient(username="a", password="b").create_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    type=AccessType.READ,
                    username='bar')
            except ResponseCodeError:
                pass

    def test_update_database_access_succeeds(self):
        exp = DatabaseAccess(type=AccessType.READ,
                             user=UserBrief(id='abdbf897-e599-4e5a-a3f0-7529884ea011', username='other'))
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                     json=exp.model_dump(),
                     status_code=202)
            # test
            response = RestClient(username="a", password="b").update_database_access(
                database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                type=AccessType.READ,
                username='bar')
            self.assertEqual(response, exp.type)

    def test_update_database_access_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                     status_code=400)
            # test
            try:
                RestClient(username="a", password="b").update_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    type=AccessType.READ,
                    username='bar')
            except MalformedError:
                pass

    def test_update_database_access_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                     status_code=403)
            # test
            try:
                RestClient(username="a", password="b").update_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    type=AccessType.READ,
                    username='bar')
            except ForbiddenError:
                pass

    def test_update_database_access_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                     status_code=404)
            # test
            try:
                RestClient(username="a", password="b").update_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    type=AccessType.READ,
                    username='bar')
            except NotExistsError:
                pass

    def test_update_database_access_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                     status_code=502)
            # test
            try:
                RestClient(username="a", password="b").update_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    type=AccessType.READ,
                    username='bar')
            except ServiceConnectionError:
                pass

    def test_update_database_access_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                     status_code=503)
            # test
            try:
                RestClient(username="a", password="b").update_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    type=AccessType.READ,
                    username='bar')
            except ServiceError:
                pass

    def test_update_database_access_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                     status_code=200)
            # test
            try:
                RestClient(username="a", password="b").update_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    type=AccessType.READ,
                    username='bar')
            except ResponseCodeError:
                pass

    def test_update_database_access_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                     status_code=404)
            # test
            try:
                RestClient().update_database_access(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                    type=AccessType.READ,
                                                    username='bar')
            except AuthenticationError:
                pass

    def test_delete_database_access_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                status_code=202)
            # test
            client = RestClient(username="a", password="b")
            RestClient(username="a", password="b").delete_database_access(
                database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                username='bar')

    def test_delete_database_access_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                status_code=400)
            # test
            try:
                RestClient(username="a", password="b").delete_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    username='bar')
            except MalformedError:
                pass

    def test_delete_database_access_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                status_code=403)
            # test
            try:
                RestClient(username="a", password="b").delete_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    username='bar')
            except ForbiddenError:
                pass

    def test_delete_database_access_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                status_code=404)
            # test
            try:
                RestClient(username="a", password="b").delete_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    username='bar')
            except NotExistsError:
                pass

    def test_delete_database_access_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                status_code=502)
            # test
            try:
                RestClient(username="a", password="b").delete_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    username='bar')
            except ServiceConnectionError:
                pass

    def test_delete_database_access_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                status_code=503)
            # test
            try:
                RestClient(username="a", password="b").delete_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    username='bar')
            except ServiceError:
                pass

    def test_delete_database_access_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                status_code=200)
            # test
            try:
                RestClient(username="a", password="b").delete_database_access(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    username='bar')
            except ResponseCodeError:
                pass

    def test_delete_database_access_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.delete(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/access/bar',
                status_code=404)
            # test
            try:
                RestClient().delete_database_access(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                    username='bar')
            except AuthenticationError:
                pass


if __name__ == "__main__":
    unittest.main()



from dbrepo.core.omlib.exceptions.dimensionexception import DimensionalException


class Dimension:

    def __init__(self, T=0, L=0, M=0, I=0, Theta=0, N=0, J=0):
        self.T = T
        self.L = L
        self.M = M
        self.I = I
        self.Theta = Theta
        self.N = N
        self.J = J

    def time_dimension_exponent(self):
        return self.T

    def length_dimension_exponent(self):
        return self.L

    def mass_dimension_exponent(self):
        return self.M

    def electric_current_dimension_exponent(self):
        return self.I

    def thermodynamic_temperature_dimension_exponent(self):
        return self.Theta

    def amount_of_substance_dimension_exponent(self):
        return self.N

    def luminous_intensity_dimension_exponent(self):
        return self.J

    def __eq__(self, other):
        return (self.T == other.T and self.L == other.L and self.M == other.M and self.I == other.I and
                self.Theta == other.Theta and self.N == other.N and self.J == other.J)

    def __str__(self):
        return f'(T={self.T}, L={self.L}, M={self.M}, I={self.I}, θ={self.Theta}, N={self.N}, J={self.J})'

    def __add__(self, other):
        if not self == other:
            raise DimensionalException("Entities of different dimensions cannot be added together. {} != {}"
                                       .format(self, other))
        return self

    def __sub__(self, other):
        if not self == other:
            raise DimensionalException("Entities of different dimensions cannot be subtracted from each other. {} != {}"
                                       .format(self, other))
        return self

    def __mul__(self, other):
        return Dimension(self.T + other.T, self.L + other.L, self.M + other.M, self.I + other.I,
                         self.Theta + other.Theta, self.N + other.N, self.J + other.J)

    def __truediv__(self, other):
        return Dimension(self.T - other.T, self.L - other.L, self.M - other.M, self.I - other.I,
                         self.Theta - other.Theta, self.N - other.N, self.J - other.J)

    @staticmethod
    def pow(base, exponent):
        if isinstance(base, Dimension):
            return Dimension(base.T * exponent, base.L * exponent, base.M * exponent, base.I * exponent,
                             base.Theta * exponent, base.N * exponent, base.J * exponent)
        else:
            return Dimension()



import json
import logging
import os
import sys

import pika
from pandas import DataFrame

from dbrepo.api.exceptions import AuthenticationError

logging.basicConfig(format='%(asctime)s %(name)-12s %(levelname)-6s %(message)s', level=logging.INFO,
                    stream=sys.stdout)


class AmqpClient:
    """
    The AmqpClient class for communicating with the DBRepo AMQP API to import data. All parameters can be set also \
    via environment variables, e.g. set endpoint with DBREPO_ENDPOINT. You can override the constructor parameters \
    with the environment variables.

    :param broker_host: The AMQP API host. Optional. Default: "localhost".
    :param broker_port: The AMQP API port. Optional. Default: 5672,
    :param broker_virtual_host: The AMQP API virtual host. Optional. Default: "dbrepo".
    :param username: The AMQP API username. Optional.
    :param password: The AMQP API password. Optional.
    """
    broker_host: str = None
    broker_port: int = 5672
    broker_virtual_host: str = None
    username: str = None
    password: str = None

    def __init__(self,
                 broker_host: str = 'localhost',
                 broker_port: int = 5672,
                 broker_virtual_host: str = 'dbrepo',
                 username: str = None,
                 password: str = None) -> None:
        self.broker_host = os.environ.get('AMQP_API_HOST', broker_host)
        self.broker_port = os.environ.get('AMQP_API_PORT', broker_port)
        if os.environ.get('AMQP_API_VIRTUAL_HOST') is not None:
            self.broker_virtual_host = os.environ.get('AMQP_API_VIRTUAL_HOST')
        else:
            self.broker_virtual_host = broker_virtual_host
        self.username = os.environ.get('AMQP_API_USERNAME', username)
        self.password = os.environ.get('AMQP_API_PASSWORD', password)

    def publish(self, routing_key: str, data: dict, exchange: str = 'dbrepo') -> None:
        """
        Publishes data to a given exchange with the given routing key with a blocking connection.

        :param routing_key: The routing key.
        :param data: The data tuple.
        :param exchange: The exchange name. Default: "dbrepo".
        """
        if self.username is None or self.password is None:
            raise AuthenticationError(f"Failed to perform request: authentication required")
        parameters = pika.ConnectionParameters(host=self.broker_host, port=self.broker_port,
                                               virtual_host=self.broker_virtual_host,
                                               credentials=pika.credentials.PlainCredentials(self.username,
                                                                                             self.password))
        connection = pika.BlockingConnection(parameters)
        channel = connection.channel()
        channel.basic_publish(exchange=exchange, routing_key=routing_key, body=json.dumps(data))
        connection.close()

    def publish_dataframe(self, routing_key: str, data: DataFrame, exchange: str = 'dbrepo') -> None:
        """
        Publishes bulk of data to a given exchange with the given routing key with a blocking connection.

        :param routing_key: The routing key.
        :param data: The dataframe
        :param exchange: The exchange name. Default: "dbrepo".
        """
        if self.username is None or self.password is None:
            raise AuthenticationError(f"Failed to perform request: authentication required")
        parameters = pika.ConnectionParameters(host=self.broker_host, port=self.broker_port,
                                               virtual_host=self.broker_virtual_host,
                                               credentials=pika.credentials.PlainCredentials(self.username,
                                                                                             self.password))
        connection = pika.BlockingConnection(parameters)
        channel = connection.channel()
        for idx, row in data.iterrows():
            body = row.to_json()
            channel.basic_publish(exchange=exchange, routing_key=routing_key, body=body)
        connection.close()



import logging
from itertools import groupby
from typing import List

import pandas
from numpy import dtype, array
from pandas import DataFrame

from dbrepo.api.dto import Subset, QueryDefinition, Database, Image, Filter, Order, CreateTableColumn, \
    CreateTableConstraints, ColumnType, SubsetColumn, FilterType, Join, Conditional
from dbrepo.api.exceptions import MalformedError


def query_to_subset(database: Database, image: Image, query: QueryDefinition) -> Subset:
    if len(query.columns) < 1:
        raise MalformedError(f'Failed to map subset: no column(s) selected')
    wrong_columns = [column for column in query.columns if len(column.split('.')) != 2]
    if query.joins is not None:
        wrong_columns += [conditional for conditional in array([join.conditionals for join in query.joins]).flatten() if
                          len(conditional.column.split('.')) != 2 or len(conditional.foreign_column.split('.')) != 2]
    if query.filters is not None:
        wrong_columns += [filter.column for filter in query.filters if
                          filter.type == FilterType.WHERE and len(filter.column.split('.')) != 2]
    if query.orders is not None:
        wrong_columns += [order.column for order in query.orders if len(order.column.split('.')) != 2]
    if len(wrong_columns) > 0:
        raise MalformedError(f'Failed to map subset: column(s) are not in table.column notion: {wrong_columns}')
    if len(query.datasources) < 1:
        raise MalformedError(f'Failed to map subset: no datasource(s) selected')
    alias_columns = [key for (key, group) in groupby(query.columns, lambda x: x) if len(list(group)) > 1]
    # columns
    select_columns = [SubsetColumn(id=column.id) for column in array(
        [table.columns for table in database.tables if
         table.internal_name in [c.split('.')[0] for c in query.columns]]).flatten() if
                      column.internal_name in [c.split('.')[1] for c in query.columns]]
    select_columns += [SubsetColumn(id=column.id) for column in array(
        [view.columns for view in database.views if
         view.internal_name in [c.split('.')[0] for c in query.columns]]).flatten() if
                       column.internal_name in [c.split('.')[1] for c in query.columns]]
    if len(select_columns) != len(query.columns):
        raise MalformedError(
            f'Failed to map subset: column(s) not found in database {len(select_columns)} != {len(query.columns)}')
    # datasources
    datasource_ids = [table.id for table in database.tables if table.internal_name in query.datasources]
    datasource_ids += [view.id for view in database.views if view.internal_name in query.datasources]
    if len(datasource_ids) != len(query.datasources):
        raise MalformedError(f'Failed to map subset: datasource(s) not found in database')
    # joins
    joins = []
    if query.joins is not None:
        for join in query.joins:
            # datasource_id
            datasource_id = [table.id for table in database.tables if table.internal_name == join.datasource]
            datasource_id += [view.id for view in database.views if view.internal_name == join.datasource]
            if len(datasource_id) != 1:
                raise MalformedError(f'Failed to map subset: join datasource {join.datasource} not found in database')
            # conditionals
            conditionals = []
            for conditional in join.conditionals:
                # column_id
                column_id = [column.id for column in array(
                    [table.columns for table in database.tables if table.internal_name == conditional.column.split('.')[
                        0]]).flatten() if column.internal_name == conditional.column.split('.')[1]]
                column_id += [column.id for column in array(
                    [view.columns for view in database.views if view.internal_name == conditional.column.split('.')[
                        0]]).flatten() if column.internal_name == conditional.column.split('.')[1]]
                if len(column_id) != 1:
                    raise MalformedError(
                        f'Failed to map subset: conditional column {conditional.column} not found in database')
                # foreign_column_id
                foreign_column_id = [column.id for column in array(
                    [table.columns for table in database.tables if table.internal_name == conditional.foreign_column.split('.')[
                        0]]).flatten() if column.internal_name == conditional.foreign_column.split('.')[1]]
                foreign_column_id += [column.id for column in array(
                    [view.columns for view in database.views if view.internal_name == conditional.foreign_column.split('.')[
                        0]]).flatten() if column.internal_name == conditional.foreign_column.split('.')[1]]
                if len(foreign_column_id) != 1:
                    raise MalformedError(
                        f'Failed to map subset: conditional column {conditional.foreign_column} not found in database')
                conditionals.append(Conditional(column_id=column_id[0], foreign_column_id=foreign_column_id[0]))
            joins.append(Join(type=join.type, datasource_id=datasource_id[0], conditionals=conditionals))
    # filters
    filters = []
    if query.filters is not None:
        for filter in query.filters:
            if filter.type != FilterType.WHERE:
                filters.append(Filter(type=filter.type))
                continue
            # column_id
            filter_column_ids = [column.id for column in array(
                [table.columns for table in database.tables if
                 table.internal_name == filter.column.split('.')[0]]).flatten()
                                 if column.internal_name == filter.column.split('.')[1]]
            filter_column_ids += [column.id for column in array(
                [view.columns for view in database.views if
                 view.internal_name == filter.column.split('.')[0]]).flatten()
                                  if column.internal_name == filter.column.split('.')[1]]
            if len(filter_column_ids) != 1:
                raise MalformedError(
                    f'Failed to map subset: filtered column name {filter.column} not found in database: {database.internal_name}')
            # operator_id
            filter_ops_ids: List[str] = [op.id for op in image.operators if op.value == filter.operator]
            if len(filter_ops_ids) != 1:
                raise MalformedError(f'Failed to map subset: filter operator {filter.operator} not found in image')
            filters.append(Filter(type=filter.type,
                                  column_id=filter_column_ids[0],
                                  operator_id=filter_ops_ids[0],
                                  value=filter.value))
    orders = []
    if query.orders is not None:
        for order in query.orders:
            # column_id
            order_column_ids = [column.id for column in array(
                [table.columns for table in database.tables if
                 table.internal_name == order.column.split('.')[0]]).flatten()
                                if column.internal_name == order.column.split('.')[1]]
            order_column_ids += [column.id for column in array(
                [view.columns for view in database.views if
                 view.internal_name == order.column.split('.')[0]]).flatten()
                                 if column.internal_name == order.column.split('.')[1]]
            if len(order_column_ids) != 1:
                raise MalformedError(
                    f'Failed to map subset: order column name {order.column} not found in database: {database.internal_name}')
            orders.append(Order(column_id=order_column_ids[0], direction=order.direction))
    return Subset(columns=select_columns, datasource_ids=datasource_ids, joins=joins, filters=filters, orders=orders)


def dataframe_to_table_definition(dataframe: DataFrame) -> ([CreateTableColumn], CreateTableConstraints):
    if dataframe.index.name is None:
        raise MalformedError(f'Failed to map dataframe: index not set')
    constraints = CreateTableConstraints(uniques=[],
                                         checks=[],
                                         foreign_keys=[],
                                         primary_key=dataframe.index.names)
    dataframe = dataframe.reset_index()
    columns = []
    for name, series in dataframe.items():
        column = CreateTableColumn(name=str(name),
                                   type=ColumnType.TEXT,
                                   null_allowed=contains_null(dataframe[name]))
        if series.dtype == dtype('float64'):
            if pandas.to_numeric(dataframe[name], errors='coerce').notnull().all():
                logging.debug(f"mapped column {name} from float64 to decimal")
                column.type = ColumnType.DECIMAL
                column.size = 40
                column.d = 20
            else:
                logging.debug(f"mapped column {name} from float64 to text")
                column.type = ColumnType.TEXT
        elif series.dtype == dtype('int64'):
            min_val = min(dataframe[name])
            max_val = max(dataframe[name])
            if 0 <= min_val <= 1 and 0 <= max_val <= 1 and 'id' not in name:
                logging.debug(f"mapped column {name} from int64 to bool")
                column.type = ColumnType.BOOL
                columns.append(column)
                continue
            logging.debug(f"mapped column {name} from int64 to bigint")
            column.type = ColumnType.BIGINT
        elif series.dtype == dtype('O'):
            try:
                pandas.to_datetime(dataframe[name], format='mixed')
                if dataframe[name].str.contains(':').any():
                    logging.debug(f"mapped column {name} from O to timestamp")
                    column.type = ColumnType.TIMESTAMP
                    columns.append(column)
                    continue
                logging.debug(f"mapped column {name} from O to date")
                column.type = ColumnType.DATE
                columns.append(column)
                continue
            except ValueError:
                pass
            max_size = max(dataframe[name].astype(str).map(len))
            if max_size <= 1:
                logging.debug(f"mapped column {name} from O to char")
                column.type = ColumnType.CHAR
                column.size = 1
            if 0 <= max_size <= 255:
                logging.debug(f"mapped column {name} from O to varchar")
                column.type = ColumnType.VARCHAR
                column.size = 255
            else:
                logging.debug(f"mapped column {name} from O to text")
                column.type = ColumnType.TEXT
        elif series.dtype == dtype('bool'):
            logging.debug(f"mapped column {name} from bool to bool")
            column.type = ColumnType.BOOL
        elif series.dtype == dtype('datetime64'):
            logging.debug(f"mapped column {name} from datetime64 to datetime")
            column.type = ColumnType.DATETIME
        else:
            logging.warning(f'default to \'text\' for column {name} and type {dtype}')
        columns.append(column)
    return columns, constraints


def contains_null(dataframe: DataFrame) -> bool:
    if '\\N' in dataframe.values:
        return True
    return dataframe.isnull().values.any()



import logging
from typing import List, Any

import requests
from jwt import jwk_from_pem, JWT
from jwt.exceptions import JWTDecodeError

from dbrepo.api.dto import ApiError
from dbrepo.core.api.dto import User


class AuthServiceClient:

    def __init__(self, endpoint: str, client_id: str, client_secret: str, jwt_public_key: str):
        self.endpoint = endpoint
        self.client_id = client_id
        self.client_secret = client_secret
        self.jwt_public_key = jwt_public_key

    def get_user_token(self, username: str, password: str) -> str:
        response = requests.post(f"{self.endpoint}/realms/dbrepo/protocol/openid-connect/token",
                                 data={
                                     "username": username,
                                     "password": password,
                                     "grant_type": "password",
                                     "client_id": self.client_id,
                                     "client_secret": self.client_secret
                                 })
        body = response.json()
        if "access_token" not in body:
            raise AssertionError("Failed to obtain user token(s)")
        return response.json()["access_token"]

    def get_user_id(self, auth_header: str | None) -> (str | None, ApiError, int):
        if auth_header is None:
            return None, None, None
        try:
            user = self.verify_jwt(auth_header.split(" ")[1])
            logging.debug(f'mapped JWT to user.id {user.id}')
            return user.id, None, None
        except JWTDecodeError as e:
            logging.error(f'Failed to decode JWT: {e}')
            if str(e) == 'JWT Expired':
                return None, ApiError(status='UNAUTHORIZED', message=f'Token expired',
                                      code='search.user.unauthorized').model_dump(), 401
            return None, ApiError(status='FORBIDDEN', message=str(e), code='search.user.forbidden').model_dump(), 403

    def get_username(self, auth_header: str | None) -> (str | None, ApiError, int):
        if auth_header is None:
            return None, None, None
        try:
            user = self.verify_jwt(auth_header.split(" ")[1])
            logging.debug(f'mapped JWT to user.username {user.username}')
            return user.username, None, None
        except JWTDecodeError as e:
            logging.error(f'Failed to decode JWT: {e}')
            if str(e) == 'JWT Expired':
                return None, ApiError(status='UNAUTHORIZED', message=f'Token expired',
                                      code='search.user.unauthorized').model_dump(), 401
            return None, ApiError(status='FORBIDDEN', message=str(e), code='search.user.forbidden').model_dump(), 403

    def verify_jwt(self, access_token: str) -> User:
        public_key = jwk_from_pem(self.jwt_public_key.encode('utf-8'))
        payload = JWT().decode(message=access_token, key=public_key, do_time_check=True)
        return User(id=payload.get('uid'), username=payload.get('preferred_username'),
                    roles=payload.get('realm_access')["roles"])

    def is_valid_token(self, token: str) -> bool | User:
        if token is None or token == "":
            return False
        try:
            return self.verify_jwt(access_token=token)
        except JWTDecodeError:
            return False

    def is_valid_password(self, username: str, password: str) -> Any:
        if username is None or username == "" or password is None or password == "":
            return False
        try:
            return self.verify_jwt(access_token=self.get_user_token(username=username, password=password))
        except AssertionError as error:
            logging.error(error)
            return False
        except requests.exceptions.ConnectionError as error:
            logging.error(f"Failed to connect to Authentication Service {error}")
            return False

    def get_user_roles(self, user: User) -> List[str]:
        return user.roles



class MockResponse:
    def __init__(self, status_code, headers=None, json_data=None):
        self.status_code = status_code
        self.headers = headers or {"Content-Type": "application/json"}
        self.json_data = json_data

    def json(self):
        return self.json_data



import unittest

import requests_mock

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import Message


class ImageUnitTest(unittest.TestCase):

    def test_get_message_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/message', json=[])
            # test
            response = RestClient().get_messages()
            self.assertEqual([], response)

    def test_get_images_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [Message(id="a456d7f0-9d42-48a8-bf5b-4ead85279e0e", type="info")]
            # mock
            mock.get('/api/v1/message', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_messages()
            self.assertEqual(exp, response)


if __name__ == "__main__":
    unittest.main()



import unittest

from dbrepo.api.dto import QueryDefinition, JoinDefinition, JoinType, ConditionalDefinition, FilterDefinition, \
    FilterType, OrderType, OrderDefinition, Database, UserBrief, Table, Constraints, PrimaryKey, TableBrief, \
    ColumnBrief, ColumnType, Column, ContainerBrief, ImageBrief, Image, Operator, DataType, SubsetColumn, Join, \
    Conditional, Filter, Order
from dbrepo.api.exceptions import MalformedError
from dbrepo.api.mapper import query_to_subset


class MapperUnitTest(unittest.TestCase):
    database = Database(
        id="6bd39359-b154-456d-b9c2-caa516a45732",
        name='test',
        owner=UserBrief(id='abdbf897-e599-4e5a-a3f0-7529884ea011', username='mweise'),
        contact=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
        exchange_name='dbrepo',
        internal_name='test_abcd',
        is_public=True,
        is_schema_public=True,
        is_dashboard_enabled=True,
        tables=[Table(id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                      name="Some Table",
                      description="Test Table",
                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                      internal_name="some_table",
                      owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                      is_versioned=True,
                      queue_name='dbrepo',
                      routing_key='dbrepo.test_database_1234.some_table',
                      is_public=True,
                      is_schema_public=True,
                      constraints=Constraints(uniques=[],
                                              foreign_keys=[],
                                              checks=[],
                                              primary_key=[PrimaryKey(id="1516310f-ecb5-4614-abe2-3b96114e1484",
                                                                      table=TableBrief(
                                                                          id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                                                          database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                          name='Some Table',
                                                                          internal_name='some_table',
                                                                          is_versioned=True,
                                                                          is_public=True,
                                                                          is_schema_public=True,
                                                                          owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16'),
                                                                      column=ColumnBrief(
                                                                          id="31a533b6-8ddf-43d6-ac6a-b9da597cb976",
                                                                          table_id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                                                          database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                          name='id',
                                                                          alias=None,
                                                                          internal_name='id',
                                                                          type=ColumnType.BIGINT))]),
                      columns=[Column(id="31a533b6-8ddf-43d6-ac6a-b9da597cb976",
                                      name="ID",
                                      ord=0,
                                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                      table_id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                      internal_name="id",
                                      type=ColumnType.BIGINT,
                                      is_null_allowed=False),
                               Column(id="85de93a8-834c-4cf4-9d34-f80ebd97e606",
                                      name="Username",
                                      ord=1,
                                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                      table_id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                      internal_name="username",
                                      type=ColumnType.VARCHAR,
                                      is_null_allowed=False)
                               ]),
                Table(id="585d421a-ad1a-4543-b661-5e32a78dd3e1",
                      name="Other Table",
                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                      internal_name="other_table",
                      owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                      is_versioned=True,
                      queue_name='dbrepo',
                      routing_key='dbrepo.test_database_1234.other_table',
                      is_public=True,
                      is_schema_public=True,
                      constraints=Constraints(uniques=[],
                                              foreign_keys=[],
                                              checks=[],
                                              primary_key=[PrimaryKey(id="3b060596-38f9-4055-8fb0-5526918f31a0",
                                                                      table=TableBrief(
                                                                          id="585d421a-ad1a-4543-b661-5e32a78dd3e1",
                                                                          database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                          name='Other Table',
                                                                          internal_name='other_table',
                                                                          is_versioned=True,
                                                                          is_public=True,
                                                                          is_schema_public=True,
                                                                          owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16'),
                                                                      column=ColumnBrief(
                                                                          id="4a2f20c3-9efd-4788-a38d-3079d198125c",
                                                                          table_id="585d421a-ad1a-4543-b661-5e32a78dd3e1",
                                                                          database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                          name='id',
                                                                          alias=None,
                                                                          internal_name='id',
                                                                          type=ColumnType.BIGINT))]),
                      columns=[Column(id="4a2f20c3-9efd-4788-a38d-3079d198125c",
                                      name="ID",
                                      ord=0,
                                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                      table_id="585d421a-ad1a-4543-b661-5e32a78dd3e1",
                                      internal_name="id",
                                      type=ColumnType.BIGINT,
                                      is_null_allowed=False),
                               Column(id="85de93a8-834c-4cf4-9d34-f80ebd97e606",
                                      name="City",
                                      ord=1,
                                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                      table_id="585d421a-ad1a-4543-b661-5e32a78dd3e1",
                                      internal_name="city",
                                      type=ColumnType.VARCHAR,
                                      is_null_allowed=False)
                               ])
                ],
        container=ContainerBrief(id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                                 name='MariaDB Galera 11.1.3',
                                 internal_name='mariadb',
                                 image=ImageBrief(id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                                  name='mariadb',
                                                  version='11.2.2',
                                                  default=True)))
    image = Image(id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                  name="mariadb",
                  version="10.11.3",
                  default=True,
                  operators=[
                      Operator(id="0917b17e-5d68-4ddf-94f6-f178f74a0dae",
                               display_name=">",
                               value=">",
                               documentation="https://mariadb.com/kb/en/greater-than/"),
                      Operator(id="e21b7b74-e7a3-4cde-9e57-b313f59c0c16",
                               display_name="<",
                               value="<",
                               documentation="https://mariadb.com/kb/en/less-than/")],
                  data_types=[
                      DataType(id="22975809-5496-4d67-9fd4-6689f0030f82",
                               display_name="SERIAL",
                               value="serial",
                               documentation="https://mariadb.com/kb/en/bigint/",
                               is_quoted=False,
                               is_buildable=True)])
    query = QueryDefinition(datasources=["some_table"],
                            columns=["other_table.city", "some_table.username"],
                            joins=[JoinDefinition(type=JoinType.INNER,
                                                  datasource="other_table",
                                                  conditionals=[ConditionalDefinition(
                                                      column="other_table.id",
                                                      foreign_column="some_table.id")])],
                            filters=[FilterDefinition(type=FilterType.WHERE,
                                                      column="some_table.id",
                                                      operator=">",
                                                      value="10"),
                                     FilterDefinition(type=FilterType.AND),
                                     FilterDefinition(type=FilterType.WHERE,
                                                      column="some_table.id",
                                                      operator="<",
                                                      value="20")],
                            orders=[OrderDefinition(column="some_table.id", direction=OrderType.DESC)])

    def test_query_to_subset_succeeds(self):
        # test
        response = query_to_subset(self.database, self.image, self.query)
        self.assertEqual([SubsetColumn(id='85de93a8-834c-4cf4-9d34-f80ebd97e606'),
                          SubsetColumn(id='85de93a8-834c-4cf4-9d34-f80ebd97e606')], response.columns)
        self.assertEqual(['029d773f-f98b-40c0-ab22-b8b1635d4fbc'], response.datasource_ids)
        self.assertEqual([Join(type=JoinType.INNER, datasource_id='585d421a-ad1a-4543-b661-5e32a78dd3e1',
                               conditionals=[Conditional(column_id='4a2f20c3-9efd-4788-a38d-3079d198125c',
                                                         foreign_column_id='31a533b6-8ddf-43d6-ac6a-b9da597cb976')])],
                         response.joins)
        self.assertEqual([Filter(type=FilterType.WHERE, column_id='31a533b6-8ddf-43d6-ac6a-b9da597cb976',
                                 operator_id='0917b17e-5d68-4ddf-94f6-f178f74a0dae', value='10'),
                          Filter(type=FilterType.AND),
                          Filter(type=FilterType.WHERE, column_id='31a533b6-8ddf-43d6-ac6a-b9da597cb976',
                                 operator_id='e21b7b74-e7a3-4cde-9e57-b313f59c0c16', value='20')], response.filters)
        self.assertEqual([Order(column_id='31a533b6-8ddf-43d6-ac6a-b9da597cb976', direction=OrderType.DESC)],
                         response.orders)

    def test_query_to_subset_column_malformed_fails(self):
        query = QueryDefinition(datasources=["other_table"],
                                columns=["other_table.city", "id"])  # <<<
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_datasource_fails(self):
        query = QueryDefinition(datasources=["i_do_not_exist"],  # <<<
                                columns=["i_do_not_exist.city", "i_do_not_exist.id"])
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_datasources_fails(self):
        query = QueryDefinition(datasources=["other_table", "i_do_not_exist"],  # <<<
                                columns=["other_table.city", "other_table.id"])
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_join_datasource_fails(self):
        query = QueryDefinition(datasources=["some_table"],
                                columns=["other_table.city", "some_table.username"],
                                joins=[JoinDefinition(type=JoinType.INNER,
                                                      datasource="i_do_not_exist",  # <<<
                                                      conditionals=[ConditionalDefinition(
                                                          column="other_table.id",
                                                          foreign_column="some_table.id")])])
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_join_conditional_column_fails(self):
        query = QueryDefinition(datasources=["some_table"],
                                columns=["other_table.city", "some_table.username"],
                                joins=[JoinDefinition(type=JoinType.INNER,
                                                      datasource="other_table",
                                                      conditionals=[ConditionalDefinition(
                                                          column="i_do_not_exist.id",  # <<<
                                                          foreign_column="some_table.id")])])
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_join_conditional_column_malformed_fails(self):
        query = QueryDefinition(datasources=["some_table"],
                                columns=["other_table.city", "some_table.username"],
                                joins=[JoinDefinition(type=JoinType.INNER,
                                                      datasource="other_table",
                                                      conditionals=[ConditionalDefinition(
                                                          column="id",  # <<<
                                                          foreign_column="some_table.id")])])
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_join_conditional_foreign_column_fails(self):
        query = QueryDefinition(datasources=["some_table"],
                                columns=["other_table.city", "some_table.username"],
                                joins=[JoinDefinition(type=JoinType.INNER,
                                                      datasource="other_table",
                                                      conditionals=[ConditionalDefinition(
                                                          column="other_table.id",
                                                          foreign_column="i_do_not_exist.id")])])  # <<<
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_join_conditional_foreign_column_malformed_fails(self):
        query = QueryDefinition(datasources=["some_table"],
                                columns=["other_table.city", "some_table.username"],
                                joins=[JoinDefinition(type=JoinType.INNER,
                                                      datasource="other_table",
                                                      conditionals=[ConditionalDefinition(
                                                          column="other_table.id",
                                                          foreign_column="id")])])  # <<<
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_join_filter_column_fails(self):
        query = QueryDefinition(datasources=["some_table"],
                                columns=["other_table.city", "some_table.username"],
                                joins=[JoinDefinition(type=JoinType.INNER,
                                                      datasource="other_table",
                                                      conditionals=[ConditionalDefinition(
                                                          column="other_table.id",
                                                          foreign_column="some_table.id")])],
                                filters=[FilterDefinition(type=FilterType.WHERE,
                                                          column="i_do_not_exist.id",  # <<<
                                                          operator=">",
                                                          value="10")])
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_join_filter_column_malformed_fails(self):
        query = QueryDefinition(datasources=["some_table"],
                                columns=["other_table.city", "some_table.username"],
                                joins=[JoinDefinition(type=JoinType.INNER,
                                                      datasource="other_table",
                                                      conditionals=[ConditionalDefinition(
                                                          column="other_table.id",
                                                          foreign_column="some_table.id")])],
                                filters=[FilterDefinition(type=FilterType.WHERE,
                                                          column="id",  # <<<
                                                          operator=">",
                                                          value="10")])
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_join_filter_operator_fails(self):
        query = QueryDefinition(datasources=["some_table"],
                                columns=["other_table.city", "some_table.username"],
                                joins=[JoinDefinition(type=JoinType.INNER,
                                                      datasource="other_table",
                                                      conditionals=[ConditionalDefinition(
                                                          column="other_table.id",
                                                          foreign_column="some_table.id")])],
                                filters=[FilterDefinition(type=FilterType.WHERE,
                                                          column="other_table.id",
                                                          operator="=",  # <<<
                                                          value="10")])
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_join_filter_type_fails(self):
        query = QueryDefinition(datasources=["some_table"],
                                columns=["other_table.city", "some_table.username"],
                                joins=[JoinDefinition(type=JoinType.INNER,
                                                      datasource="other_table",
                                                      conditionals=[ConditionalDefinition(
                                                          column="other_table.id",
                                                          foreign_column="some_table.id")])],
                                filters=[FilterDefinition(type=FilterType.WHERE,
                                                          column="other_table.id",
                                                          operator="<",
                                                          value="10"),
                                         FilterDefinition(type=FilterType.AND)])  # <<<
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_order_column_fails(self):
        query = QueryDefinition(datasources=["some_table"],
                                columns=["some_table.username"],
                                orders=[OrderDefinition(column="some_table.id")])  # <<<
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass

    def test_query_to_subset_order_column_malformed_fails(self):
        query = QueryDefinition(datasources=["some_table"],
                                columns=["some_table.username"],
                                orders=[OrderDefinition(column="id")])  # <<<
        # test
        try:
            query_to_subset(self.database, self.image, query)
        except MalformedError:
            pass



import rdflib
from rdflib import URIRef

from dbrepo.core.omlib.dimension import Dimension
from dbrepo.core.omlib.scale import Scale
from dbrepo.core.omlib.unit import Prefix, Unit


class OM_IDS:
    NAMESPACE = 'http://www.ontology-of-units-of-measure.org/resource/om-2/'


class SI:
    SYSTEM_OF_UNITS = str(OM_IDS.NAMESPACE + 'InternationalSystemOfUnits')  # WARNING If you change "SI" also change in constants.py

    # SI Prefixes
    YOTTA = Prefix('yotta', 'Y', 1e24, OM_IDS.NAMESPACE + 'yotta')
    ZETTA = Prefix('zetta', 'Z', 1e21, OM_IDS.NAMESPACE + 'zetta')
    EXA = Prefix('exa', 'E', 1e18, OM_IDS.NAMESPACE + 'exa')
    PETA = Prefix('peta', 'P', 1e15, OM_IDS.NAMESPACE + 'peta')
    TERA = Prefix('tera', 'T', 1e12, OM_IDS.NAMESPACE + 'tera')
    GIGA = Prefix('giga', 'G', 1e9, OM_IDS.NAMESPACE + 'giga')
    MEGA = Prefix('mega', 'M', 1e6, OM_IDS.NAMESPACE + 'mega')
    KILO = Prefix('kilo', 'k', 1e3, OM_IDS.NAMESPACE + 'kilo')
    HECTO = Prefix('hecto', 'h', 1e2, OM_IDS.NAMESPACE + 'hecto')
    DECA = Prefix('deca', 'da', 1e1, OM_IDS.NAMESPACE + 'deca')
    DECI = Prefix('deci', 'd', 1e-1, OM_IDS.NAMESPACE + 'deci')
    CENTI = Prefix('centi', 'c', 1e-2, OM_IDS.NAMESPACE + 'centi')
    MILLI = Prefix('milli', 'm', 1e-3, OM_IDS.NAMESPACE + 'milli')
    MICRO = Prefix('micro', 'μ', 1e-6, OM_IDS.NAMESPACE + 'micro')
    NANO = Prefix('nano', 'n', 1e-9, OM_IDS.NAMESPACE + 'nano')
    PICO = Prefix('pico', 'p', 1e-12, OM_IDS.NAMESPACE + 'pico')
    FEMTO = Prefix('femto', 'f', 1e-15, OM_IDS.NAMESPACE + 'femto')
    ATTO = Prefix('atto', 'a', 1e-18, OM_IDS.NAMESPACE + 'atto')
    ZEPTO = Prefix('zepto', 'z', 1e-21, OM_IDS.NAMESPACE + 'zepto')
    YOCTO = Prefix('yocto', 'y', 1e-24, OM_IDS.NAMESPACE + 'yocto')

    # SI Base Units
    SECOND = Unit.get_singular_unit('second', 's', Dimension(1, 0, 0, 0, 0, 0, 0),
                                    identifier=OM_IDS.NAMESPACE + 'second-Time',
                                    system_of_units=SYSTEM_OF_UNITS, is_base_unit=True)
    METRE = Unit.get_singular_unit('metre', 'm', Dimension(0, 1, 0, 0, 0, 0, 0), identifier=OM_IDS.NAMESPACE + 'metre',
                                   system_of_units=SYSTEM_OF_UNITS, is_base_unit=True)
    GRAM = Unit.get_singular_unit('gram', 'g', Dimension(0, 0, 1, 0, 0, 0, 0), identifier=OM_IDS.NAMESPACE + 'gram')
    KILOGRAM = Unit.get_prefixed_unit(KILO, GRAM, identifier=OM_IDS.NAMESPACE + 'kilogram',
                                      system_of_units=SYSTEM_OF_UNITS,
                                      is_base_unit=True)
    AMPERE = Unit.get_singular_unit('ampere', 'A', Dimension(0, 0, 0, 1, 0, 0, 0),
                                    identifier=OM_IDS.NAMESPACE + 'ampere',
                                    system_of_units=SYSTEM_OF_UNITS, is_base_unit=True)
    KELVIN = Unit.get_singular_unit('kelvin', 'K', Dimension(0, 0, 0, 0, 1, 0, 0),
                                    identifier=OM_IDS.NAMESPACE + 'kelvin',
                                    system_of_units=SYSTEM_OF_UNITS, is_base_unit=True)
    MOLE = Unit.get_singular_unit('mole', 'mol', Dimension(0, 0, 0, 0, 0, 1, 0),
                                  identifier=OM_IDS.NAMESPACE + 'mole',
                                  system_of_units=SYSTEM_OF_UNITS, is_base_unit=True)
    CANDELA = Unit.get_singular_unit('candela', 'cd', Dimension(0, 0, 0, 0, 0, 0, 1),
                                     identifier=OM_IDS.NAMESPACE + 'candela', system_of_units=SYSTEM_OF_UNITS,
                                     is_base_unit=True)


class IEC:
    KIBI = Prefix('kibi', 'Ki', pow(2, 10), OM_IDS.NAMESPACE + 'kibi')
    MEBI = Prefix('mebi', 'Mi', pow(2, 20), OM_IDS.NAMESPACE + 'mebi')
    GIBI = Prefix('gibi', 'Gi', pow(2, 30), OM_IDS.NAMESPACE + 'gibi')
    TEBI = Prefix('tebi', 'Ti', pow(2, 40), OM_IDS.NAMESPACE + 'tebi')
    PEBI = Prefix('pebi', 'Pi', pow(2, 50), OM_IDS.NAMESPACE + 'pebi')
    EXBI = Prefix('exbi', 'Ei', pow(2, 60), OM_IDS.NAMESPACE + 'exbi')
    ZEBI = Prefix('zebi', 'Zi', pow(2, 70), OM_IDS.NAMESPACE + 'zebi')
    YOBI = Prefix('yobi', 'Yi', pow(2, 80), OM_IDS.NAMESPACE + 'yobi')


class JEDEC:
    KILO = Prefix('kilo', 'k', pow(2, 10), OM_IDS.NAMESPACE + 'jedec-kilo')
    MEGA = Prefix('mega', 'M', pow(2, 20), OM_IDS.NAMESPACE + 'jedec-mega')
    GIGA = Prefix('giga', 'G', pow(2, 30), OM_IDS.NAMESPACE + 'jedec-giga')







import math

from rdflib import URIRef

from dbrepo.core.omlib.exceptions.dimensionexception import DimensionalException
from dbrepo.core.omlib.exceptions.unitconversionexception import UnitConversionException
from dbrepo.core.omlib.exceptions.unitidentityexception import UnitIdentityException
from dbrepo.core.omlib.dimension import Dimension
from dbrepo.core.omlib.thing import SymbolThing


class Unit(SymbolThing):
    _units = []

    @staticmethod
    def with_label(label):
        get_units = []
        for unit in Unit._units:
            for u_label in unit.all_labels():
                if str(u_label) == label:
                    get_units.append(unit)
                    break
        return get_units

    @staticmethod
    def with_symbol(symbol):
        get_units = []
        for unit in Unit._units:
            for u_symbol in unit.all_symbols():
                if str(u_symbol) == symbol:
                    get_units.append(unit)
        return get_units

    @staticmethod
    def with_identifier(identifier):
        for unit in Unit._units:
            if str(unit.identifier) == str(identifier):
                return unit
        return None

    @staticmethod
    def with_prefix(prefix, base):
        test_units = Unit.with_dimensions(base.dimensions)
        for unit in test_units:
            if isinstance(unit, PrefixedUnit):
                if unit.baseUnit == base and unit.prefix == prefix:
                    return unit
        return None

    @staticmethod
    def with_multiplication(multiplier, multiplicand):
        new_dimensions = multiplier.dimensions * multiplicand.dimensions
        test_units = Unit.with_dimensions(new_dimensions)
        for unit in test_units:
            if isinstance(unit, UnitMultiplication):
                if unit.multiplier == multiplier and unit.multiplicand == multiplicand:
                    return unit
        return None

    @staticmethod
    def with_division(numerator, denominator):
        new_dimensions = numerator.dimensions / denominator.dimensions
        test_units = Unit.with_dimensions(new_dimensions)
        for unit in test_units:
            if isinstance(unit, UnitDivision):
                if unit.numerator == numerator and unit.denominator == denominator:
                    return unit
        return None

    @staticmethod
    def with_exponentiation(base, exponent):
        new_dimensions = Dimension.pow(base.dimensions, exponent)
        test_units = Unit.with_dimensions(new_dimensions)
        for unit in test_units:
            if isinstance(unit, UnitExponentiation):
                if unit.base == base and unit.exponent == exponent:
                    return unit
        return None

    @staticmethod
    def with_dimensions(dimensions, in_system_of_units=None):
        if isinstance(dimensions, Dimension):
            get_units = []
            for test_unit in Unit._units:
                if test_unit.dimensions == dimensions:
                    if in_system_of_units is None or test_unit.systemOfUnits == in_system_of_units:
                        get_units.append(test_unit)
            return get_units
        return None

    @staticmethod
    def get_base_units(for_unit, in_system_of_units=str('http://www.ontology-of-units-of-measure.org/resource/om-2/InternationalSystemOfUnits')):  # WARNING If you change "SI" also change in constants.py
        if isinstance(for_unit, SingularUnit) or isinstance(for_unit, PrefixedUnit) \
                or isinstance(for_unit, UnitMultiple):
            units = Unit.with_dimensions(for_unit.dimensions)
            for unit in units:
                if unit.isBaseUnit and unit.systemOfUnits == in_system_of_units:
                    return unit
        if isinstance(for_unit, UnitMultiplication):
            base_multiplier = Unit.get_base_units(for_unit.multiplier, in_system_of_units)
            base_multiplicand = Unit.get_base_units(for_unit.multiplicand, in_system_of_units)
            if base_multiplier is not None and base_multiplicand is not None:
                unit = Unit.get_unit_multiplication(base_multiplier, base_multiplicand,
                                                    system_of_units=in_system_of_units)
                return unit
        if isinstance(for_unit, UnitDivision):
            base_numerator = Unit.get_base_units(for_unit.numerator, in_system_of_units)
            base_denominator = Unit.get_base_units(for_unit.denominator, in_system_of_units)
            if base_numerator is not None and base_denominator is not None:
                unit = Unit.get_unit_division(base_numerator, base_denominator, system_of_units=in_system_of_units)
                return unit
        if isinstance(for_unit, UnitExponentiation):
            base_base = Unit.get_base_units(for_unit.base, in_system_of_units)
            if base_base is not None:
                unit = Unit.get_unit_exponentiation(base_base, for_unit.exponent, system_of_units=in_system_of_units)
                return unit
        raise UnitIdentityException("Cannot find a base unit for {} in the system of units: {}"
                                    .format(for_unit, in_system_of_units))

    @staticmethod
    def __add_when_not_duplicate(unit):
        result_unit = Unit.with_identifier(unit.identifier)
        dim_units = Unit.with_dimensions(unit.dimensions)
        if result_unit is None:
            for test_unit in dim_units:
                try:
                    conversion_factor = abs(Unit.conversion_factor(unit, test_unit) - 1.0)
                    if conversion_factor < 0.0000001:
                        result_unit = test_unit
                        break
                except UnitConversionException as error:
                    pass
        if result_unit is not None:
            if unit.label() is not None or unit.symbol() is not None:
                all_labels = result_unit.all_labels()
                found = False
                str_unit_label = str(unit.label())
                for label in all_labels:
                    if str(label) == str_unit_label:
                        found = True
                        break
                if not found:
                    result_unit = None
        if result_unit is None:
            result_unit = unit
            Unit._units.append(unit)
        return result_unit

    @staticmethod
    def simplified_compound_unit(unit):
        # creates a simplified unit (e.g. kg.mg/g becomes kg), conversion to this unit is however needed afterwards.
        if isinstance(unit, CompoundUnit):
            if isinstance(unit, UnitDivision):
                if not isinstance(unit.numerator, CompoundUnit) and not isinstance(unit.denominator, CompoundUnit) \
                        and unit.numerator.dimensions == unit.denominator.dimensions:
                    return unit

            exponents = Unit.__exponents_of_units_of_same_dimensions(unit)
            result = Unit.__create_unit_from_exponents(exponents)
            result.identifier = unit.identifier
            return result
        return unit

    @staticmethod
    def __exponents_of_units_of_same_dimensions(unit):
        # takes together the exponents of units that have the same dimensions.
        # So if you have kg.g this will result in exponent 2 for kg.
        reduced = []
        if isinstance(unit, CompoundUnit):
            bue = unit.get_units_exponents()
            for exp in bue:
                found = False
                for red in reduced:
                    if exp[0].dimensions == red[0].dimensions:
                        red[1] = red[1] + exp[1]
                        found = True
                if not found:
                    reduced.append([exp[0], exp[1]])
            return reduced
        reduced.append([unit, 1.0])
        return reduced

    @staticmethod
    def __create_unit_from_exponents(exponents):
        numerator = None
        denominator = None
        units_in_numerator = 0
        units_in_denominator = 0
        for red in exponents:
            if red[1] > 0:
                units_in_numerator += 1
            if red[1] < 0:
                units_in_denominator += 1
        for red in exponents:
            if red[1] > 0:
                unit_exp = red[0]
                if units_in_numerator <= 1 or str(
                        unit_exp.identifier) != "http://www.ontology-of-units-of-measure.org/resource/om-2/one":
                    if red[1] > 1:
                        unit_exp = UnitExponentiation(red[0], red[1])
                    if numerator is None:
                        numerator = unit_exp
                    else:
                        numerator = UnitMultiplication(numerator, unit_exp)
            if red[1] < 0:
                unit_exp = red[0]
                if unit_exp.identifier != str("http://www.ontology-of-units-of-measure.org/resource/om-2/one"):
                    if red[1] < -1:
                        unit_exp = UnitExponentiation(red[0], -red[1])
                    if denominator is None:
                        denominator = unit_exp
                    else:
                        denominator = UnitMultiplication(denominator, unit_exp)
        if numerator is None:
            numerator = Unit.get_singular_unit('one', '', Dimension())
        if denominator is None:
            result = numerator
        else:
            result = UnitDivision(numerator, denominator)
        return result

    @staticmethod
    def reduce_unit(unit):
        result = unit
        if isinstance(unit, CompoundUnit):
            reduced = []
            bue = unit.get_units_exponents()
            for exponent in bue:
                found = False
                for red in reduced:
                    if exponent[0] == red[0]:
                        found = True
                        red[1] = red[1] + exponent[1]
                if not found:
                    reduced.append(exponent)
            reduced.sort(key=Unit.cmp_to_key(Unit.__exponents_cmp))
            if isinstance(unit, UnitDivision) \
                    and not isinstance(unit.numerator, CompoundUnit) \
                    and not isinstance(unit.denominator, CompoundUnit) \
                    and unit.numerator.dimensions == unit.denominator.dimensions:
                return unit
            result = Unit.__create_unit_from_exponents(reduced)
            result.identifier = unit.identifier
        return result

    @staticmethod
    def __exponents_cmp(a, b):
        if a[1] > 0 and b[1] < 0:
            return -1
        if a[1] < 0 and b[1] > 0:
            return 1
        if a[1] > 0 and b[1] < a[1]:
            return 1
        if 0 < a[1] < b[1]:
            return -1
        if 0 > a[1] > b[1]:
            return -1
        if a[1] < 0 and b[1] > a[1]:
            return 1
        return 0

    @staticmethod
    def cmp_to_key(mycmp):
        'Convert a cmp= function into a key= function'

        class K:
            def __init__(self, obj, *args):
                self.obj = obj

            def __lt__(self, other):
                return mycmp(self.obj, other.obj) < 0

            def __gt__(self, other):
                return mycmp(self.obj, other.obj) > 0

            def __eq__(self, other):
                return mycmp(self.obj, other.obj) == 0

            def __le__(self, other):
                return mycmp(self.obj, other.obj) <= 0

            def __ge__(self, other):
                return mycmp(self.obj, other.obj) >= 0

            def __ne__(self, other):
                return mycmp(self.obj, other.obj) != 0

        return K

    @staticmethod
    def conversion_factor(from_unit, to_unit):
        can_convert = Unit.can_convert(from_unit, to_unit)
        if not can_convert:
            raise DimensionalException("A unit with dimensions {} cannot be converted to a unit with dimensions {}."
                                       .format(from_unit.dimensions, to_unit.dimensions))
        from_base_units_exponents = from_unit.get_base_units_exponents()
        to_base_units_exponents = to_unit.get_base_units_exponents()
        from_norm = Unit.__normalise_base_units_exponents(from_base_units_exponents)
        to_norm = Unit.__normalise_base_units_exponents(to_base_units_exponents)
        if not Unit.__check_same_units_from_normalised(from_norm, to_norm):
            raise UnitConversionException("Cannot convert from {} to {} as they do not have a common ancestor unit."
                                          .format(from_unit, to_unit))
        from_factor = Unit.__factor_from_normalised(from_norm)
        to_factor = Unit.__factor_from_normalised(to_norm)
        factor = from_factor / to_factor
        return factor

    @staticmethod
    def __normalise_base_units_exponents(base_unit_exponents):
        unit_dict = {}
        for exponent in base_unit_exponents:
            entry = [0, 1.0]
            if exponent[0].identifier in unit_dict:
                entry = unit_dict[exponent[0].identifier]
            entry[0] = entry[0] + exponent[1]
            entry[1] = entry[1] * pow(exponent[2], exponent[1])
            unit_dict[exponent[0].identifier] = entry
        return unit_dict

    @staticmethod
    def __factor_from_normalised(normalised):
        factor = 1.0
        for entry in normalised:
            factor = factor * normalised.get(entry)[1]
        return factor

    @staticmethod
    def __check_same_units_from_normalised(normalised_unit1, normalised_unit2):
        keys1 = normalised_unit1.keys()
        for ident in keys1:
            exp_1 = normalised_unit1.get(ident)
            exp_2 = normalised_unit2.get(ident)
            if exp_2 is None or exp_1[0] != exp_2[0]:
                return False
        keys2 = normalised_unit2.keys()
        for ident in keys2:
            exp_1 = normalised_unit1.get(ident)
            exp_2 = normalised_unit2.get(ident)
            if exp_1 is None or exp_1[0] != exp_2[0]:
                return False
        return True

    @staticmethod
    def print_base_units_exponents(unit):
        base_units_exponents = unit.get_base_units_exponents()
        bue_string = ""
        i = 0
        for exponent in base_units_exponents:
            bue_string += f'[{str(exponent[0].symbol()), exponent[1], exponent[2]}]'
            i = i + 1
            if i < len(base_units_exponents):
                bue_string += ', '

        print('BASE UNIT EXPONENTS: {}: [{}]'.format(unit.symbol(), bue_string))

    @staticmethod
    def can_convert(from_unit, to_unit):
        if isinstance(from_unit, Unit) and isinstance(to_unit, Unit):
            if from_unit.dimensions == to_unit.dimensions:
                return True
            return False
        if not isinstance(from_unit, Unit) and not isinstance(to_unit, Unit):
            raise ValueError("Both arguments of can_convert are required to be of type Unit.")
        if not isinstance(from_unit, Unit):
            raise ValueError("The first argument of can_convert is required to be of type Unit.")
        if not isinstance(to_unit, Unit):
            raise ValueError("The second argument of can_convert is required to be of type Unit.")

    @staticmethod
    def get_singular_unit(label, symbol, dimensions=Dimension(), base_unit=None, factor=1.0, identifier=None,
                          cache=True, system_of_units=None, is_base_unit=False):
        unit = None
        if identifier is not None:
            test_unit = Unit.with_identifier(identifier)
            if test_unit is not None:
                if not isinstance(test_unit, SingularUnit):
                    raise UnitIdentityException("The identifier for a SingularUnit has been used earlier"
                                                " for another type of unit")
                else:
                    if base_unit is not None and base_unit != test_unit.baseUnit:
                        base_identifier = str(base_unit.identifier)
                        test_identifier = str(test_unit.baseUnit.identifier)
                        base_ok = False
                        factor_ok = False
                        if base_identifier == test_identifier:
                            base_ok = True
                            if factor == test_unit.factor:
                                factor_ok = True
                        else:
                            can_convert_base = Unit.can_convert(base_unit, test_unit.baseUnit)
                            if can_convert_base:
                                base_ok = True
                                conv_factor = Unit.conversion_factor(base_unit, test_unit.baseUnit)
                                tot_factor = conv_factor * factor
                                log_tot_factor = math.log10(tot_factor)
                                log_factor = math.log10(test_unit.factor)
                                diff = abs(log_tot_factor - log_factor)
                                if diff < 0.00001:
                                    factor_ok = True
                        if not base_ok:
                            raise UnitIdentityException("The requested SingularUnit uses a different base unit"
                                                        " as the earlier defined unit with the same identifier.")
                        if not factor_ok:
                            raise UnitIdentityException("The requested SingularUnit uses a different conversion factor"
                                                        " as the earlier defined unit with the same identifier.")
                if label is not None:
                    test_unit.add_preferred_label(label)
                if symbol is not None:
                    test_unit.add_symbol(symbol)
                if system_of_units is not None and test_unit.systemOfUnits is None:
                    test_unit.systemOfUnits = system_of_units
                if is_base_unit:
                    test_unit.isBaseUnit = is_base_unit
                return test_unit
        if unit is None:
            if base_unit is not None and base_unit.systemOfUnits is not None and system_of_units is None \
                    and factor == 1.0:
                system_of_units = base_unit.systemOfUnits
            unit = SingularUnit(label, symbol, dimensions, base_unit, factor, identifier, system_of_units, is_base_unit)
            if cache:
                unit = Unit.__add_when_not_duplicate(unit)
        return unit

    @staticmethod
    def get_prefixed_unit(prefix, base_unit, identifier=None, cache=True, system_of_units=None, is_base_unit=False):
        unit = None
        if isinstance(prefix, str):
            prefix = Prefix.with_identifier(prefix)
        if isinstance(prefix, URIRef):
            prefix = Prefix.with_identifier(prefix.value)
        if identifier is not None:
            test_unit = Unit.with_identifier(identifier)
            if test_unit is not None:
                if not isinstance(test_unit, PrefixedUnit):
                    raise UnitIdentityException("The identifier for a PrefixedUnit has been used earlier"
                                                " for another type of unit")
                else:
                    base_identifier = str(base_unit.identifier)
                    test_identifier = str(test_unit.baseUnit.identifier)
                    if base_identifier != test_identifier:
                        raise UnitIdentityException("The requested PrefixedUnit uses a different base unit"
                                                    " as the earlier defined unit with the same identifier.")
                    if str(prefix.identifier) != str(test_unit.prefix.identifier):
                        raise UnitIdentityException("The requested PrefixedUnit uses a different prefix"
                                                    " as the earlier defined unit with the same identifier.")
                if system_of_units is not None and test_unit.systemOfUnits is None:
                    test_unit.systemOfUnits = system_of_units
                if base_unit.systemOfUnits is not None and test_unit.systemOfUnits is None:
                    test_unit.systemOfUnits = base_unit.systemOfUnits
                if is_base_unit:
                    test_unit.isBaseUnit = is_base_unit
                return test_unit
        if unit is None:
            test_unit = Unit.with_prefix(prefix, base_unit)
            if test_unit is not None:
                if identifier is not None and not isinstance(test_unit.identifier, URIRef):
                    test_unit.identifier = URIRef(identifier)
                if system_of_units is not None and test_unit.systemOfUnits is None:
                    test_unit.systemOfUnits = system_of_units
                if base_unit.systemOfUnits is not None and test_unit.systemOfUnits is None:
                    test_unit.systemOfUnits = base_unit.systemOfUnits
                if is_base_unit and not test_unit.isBaseUnit:
                    test_unit.isBaseUnit = is_base_unit
                return test_unit
        if unit is None:
            if system_of_units is None:
                system_of_units = base_unit.systemOfUnits
            unit = PrefixedUnit(prefix, base_unit, identifier, system_of_units=system_of_units,
                                is_base_unit=is_base_unit)
            if cache:
                unit = Unit.__add_when_not_duplicate(unit)
        return unit

    @staticmethod
    def get_unit_multiple(base_unit, factor=1.0, identifier=None, label=None, symbol=None, cache=True,
                          system_of_units=None):
        unit = None
        if identifier is not None:
            test_unit = Unit.with_identifier(identifier)
            if test_unit is not None:
                if not isinstance(test_unit, UnitMultiple):
                    raise UnitIdentityException("The identifier for a UnitMultiple has been used earlier"
                                                " for another type of unit")
                else:
                    base_identifier = str(base_unit.identifier)
                    test_identifier = str(test_unit.baseUnit.identifier)
                    if base_identifier != test_identifier:
                        raise UnitIdentityException("The requested UnitMultiple uses a different base unit"
                                                    " as the earlier defined unit with the same identifier.")
                if label is not None:
                    test_unit.add_preferred_label(label)
                if symbol is not None:
                    test_unit.add_symbol(symbol)
                if system_of_units is not None and test_unit.systemOfUnits is None:
                    test_unit.systemOfUnits = system_of_units
                if base_unit.systemOfUnits is not None and test_unit.systemOfUnits is None:
                    test_unit.systemOfUnits = base_unit.systemOfUnits
                return test_unit
        if unit is None:
            unit = UnitMultiple(base_unit, factor, identifier, label, symbol, system_of_units)
            if system_of_units is None:
                system_of_units = base_unit.systemOfUnits
            if cache:
                unit = Unit.__add_when_not_duplicate(unit)
        return unit

    @staticmethod
    def get_unit_multiplication(multiplier, multiplicand, identifier=None, cache=True,
                                system_of_units=None):
        if identifier is not None:
            for test_unit in Unit._units:
                if str(test_unit.identifier) == str(identifier):
                    if not isinstance(test_unit, UnitMultiplication):
                        raise UnitIdentityException("The identifier for a UnitMultiplication has been used earlier"
                                                    " for another type of unit")
                    else:
                        if str(multiplier.identifier) != str(test_unit.multiplier.identifier) or \
                                str(multiplicand.identifier) != str(test_unit.multiplicand.identifier):
                            raise UnitIdentityException("The requested UnitMultiplication uses a different pair of"
                                                        " units as multiplier and multiplicand as the earlier defined"
                                                        " unit with the same identifier.")
                    if system_of_units is not None and test_unit.systemOfUnits is None:
                        test_unit.systemOfUnits = system_of_units
                    if multiplier.systemOfUnits is not None and multiplicand.systemOfUnits is not None \
                            and multiplier.systemOfUnits == multiplicand.systemOfUnits \
                            and test_unit.systemOfUnits is None:
                        test_unit.systemOfUnits = multiplier.systemOfUnits
                    return test_unit
        unit = Unit.with_multiplication(multiplier, multiplicand)
        if unit is not None:
            if identifier is not None:
                unit.identifier = identifier
                if system_of_units is not None and unit.systemOfUnits is None:
                    unit.systemOfUnits = system_of_units
                if multiplier.systemOfUnits is not None and multiplicand.systemOfUnits is not None \
                        and multiplier.systemOfUnits == multiplicand.systemOfUnits and unit.systemOfUnits is None:
                    unit.systemOfUnits = multiplier.systemOfUnits
        if unit is None:
            for unit in Unit._units:
                if isinstance(unit, UnitMultiplication):
                    if str(multiplier.identifier) == str(unit.multiplier.identifier) and \
                            str(multiplicand.identifier) == str(unit.multiplicand.identifier):
                        if system_of_units is not None:
                            unit.systemOfUnits = system_of_units
                        return unit
            if system_of_units is None and multiplier.systemOfUnits is not None and \
                    multiplicand.systemOfUnits is not None and multiplicand.systemOfUnits == multiplier.systemOfUnits:
                system_of_units = multiplier.systemOfUnits
            unit = UnitMultiplication(multiplier, multiplicand, identifier, system_of_units)
            unit = Unit.reduce_unit(unit)
            if cache:
                unit = Unit.__add_when_not_duplicate(unit)
        return unit

    @staticmethod
    def get_unit_division(numerator, denominator, identifier=None, cache=True, system_of_units=None):
        if identifier is not None:
            for test_unit in Unit._units:
                if str(test_unit.identifier) == str(identifier):
                    if not isinstance(test_unit, UnitDivision):
                        raise UnitIdentityException("The identifier for a UnitDivision has been used earlier"
                                                    " for another type of unit")
                    else:
                        if str(numerator.identifier) != str(test_unit.numerator.identifier) or \
                                str(denominator.identifier) != str(test_unit.denominator.identifier):
                            raise UnitIdentityException("The requested UnitDivision uses a different pair of"
                                                        " units as numerator and denominator as the earlier "
                                                        "defined unit with the same identifier.")
                    if system_of_units is not None and test_unit.systemOfUnits is None:
                        test_unit.systemOfUnits = system_of_units
                    if numerator.systemOfUnits is not None and denominator.systemOfUnits is not None \
                            and numerator.systemOfUnits == denominator.systemOfUnits \
                            and test_unit.systemOfUnits is None:
                        test_unit.systemOfUnits = numerator.systemOfUnits
                    return test_unit
        unit = Unit.with_division(numerator, denominator)
        if unit is not None:
            if identifier is not None:
                unit.identifier = identifier
            if system_of_units is not None:
                unit.systemOfUnits = system_of_units
            if numerator.systemOfUnits is not None and denominator.systemOfUnits is not None \
                    and numerator.systemOfUnits == denominator.systemOfUnits and unit.systemOfUnits is None:
                unit.systemOfUnits = numerator.systemOfUnits
        if unit is None:
            for unit in Unit._units:
                if isinstance(unit, UnitDivision):
                    if str(numerator.identifier) == str(unit.numerator.identifier) and \
                            str(denominator.identifier) == str(unit.denominator.identifier):
                        if system_of_units is not None:
                            unit.systemOfUnits = system_of_units
                        return unit
            if system_of_units is None and denominator.systemOfUnits is not None and \
                    numerator.systemOfUnits is not None and numerator.systemOfUnits == denominator.systemOfUnits:
                system_of_units = numerator.systemOfUnits
            unit = UnitDivision(numerator, denominator, identifier, system_of_units)
            unit = Unit.reduce_unit(unit)
            if cache:
                unit = Unit.__add_when_not_duplicate(unit)
        return unit

    @staticmethod
    def get_unit_exponentiation(base, exponent, identifier=None, cache=True, system_of_units=None):
        unit = None
        if identifier is not None:
            for test_unit in Unit._units:
                if str(test_unit.identifier) == str(identifier):
                    if not isinstance(test_unit, UnitExponentiation):
                        raise UnitIdentityException("The identifier for a UnitExponentiation has been used earlier"
                                                    " for another type of unit")
                    else:
                        if str(base.identifier) != str(test_unit.base.identifier):
                            raise UnitIdentityException("The requested UnitExponentiation uses a different unit "
                                                        "as base as the earlier defined unit with the same identifier.")
                        if exponent != test_unit.exponent:
                            raise UnitIdentityException("The requested UnitExponentiation uses a different exponent as"
                                                        " the earlier defined unit with the same identifier.")
                    if system_of_units is not None and test_unit.systemOfUnits is None:
                        test_unit.systemOfUnits = system_of_units
                    if base.systemOfUnits is not None and test_unit.systemOfUnits is None:
                        test_unit.systemOfUnits = base.systemOfUnits
                    return test_unit
        unit = Unit.with_exponentiation(base, exponent)
        if unit is not None:
            if identifier is not None:
                unit.identifier = identifier
            if system_of_units is not None:
                unit.systemOfUnits = system_of_units
        if unit is None:
            for unit in Unit._units:
                if isinstance(unit, UnitExponentiation):
                    if str(base.identifier) == str(unit.base.identifier) and \
                            exponent == unit.exponent:
                        if system_of_units is not None:
                            unit.systemOfUnits = system_of_units
                        return unit
            if system_of_units is None and base.systemOfUnits is not None:
                system_of_units = base.systemOfUnits
            unit = UnitExponentiation(base, exponent, identifier, system_of_units)
            unit = Unit.reduce_unit(unit)
            if cache:
                unit = Unit.__add_when_not_duplicate(unit)
        return unit

    def __init__(self, label=None, symbol=None, dimensions=Dimension(), identifier=None, system_of_units=None,
                 is_base_unit=False):
        super().__init__(label, symbol, identifier)
        self.isBaseUnit = is_base_unit
        self.systemOfUnits = system_of_units
        self.dimensions = dimensions

    def __str__(self):
        return f'{self.label()}\t{self.symbol()}\t<{self.identifier}>  dim: {self.dimensions}'

    def __eq__(self, other):
        if isinstance(other, Unit):
            return str(self.identifier) == str(other.identifier)
        return False

    def __ne__(self, other):
        return not (self == other)


class Prefix(object):
    _prefixes = dict()

    @staticmethod
    def with_identifier(identifier):
        identifier_string = str(identifier)
        if identifier_string in Prefix._prefixes:
            return Prefix._prefixes[identifier_string]
        return None

    def __init__(self, name, symbol, factor, identifier):
        self.name = name
        self.symbol = symbol
        self.factor = factor
        self.identifier = identifier
        Prefix._prefixes[str(identifier)] = self

    def __eq__(self, other):
        if isinstance(other, Prefix):
            return str(self.identifier) == str(other.identifier)
        return False

    def __ne__(self, other):
        return not (self == other)


class SingularUnit(Unit):

    def __init__(self, label, symbol, dimensions=Dimension(), base_unit=None, factor=1.0, identifier=None,
                 system_of_units=None, is_base_unit=False):
        if base_unit is not None:
            dimensions = base_unit.dimensions
        if system_of_units is None and not is_base_unit and base_unit is not None and factor == 1.0:
            system_of_units = base_unit.systemOfUnits
        super().__init__(label, symbol, dimensions, identifier, system_of_units=system_of_units,
                         is_base_unit=is_base_unit)
        self.factor = factor
        self.baseUnit = base_unit

    def __eq__(self, other):
        if isinstance(other, SingularUnit):
            if not isinstance(self.identifier, URIRef) or not isinstance(other.identifier, URIRef):
                if self.baseUnit is None and other.baseUnit == self and other.factor == self.factor:
                    return True
                if other.baseUnit is None and self.baseUnit == other and other.factor == self.factor:
                    return True
                if other.baseUnit == self.baseUnit and other.factor == self.factor:
                    return True
            else:
                return str(self.identifier) == str(other.identifier)
        return False

    def get_units_exponents(self):
        return [[self, 1, 1.0]]

    def get_base_units_exponents(self):
        if self.baseUnit is None:
            return [[self, 1, 1.0]]
        base_base = self.baseUnit.get_base_units_exponents()
        result = []
        factor_added = False
        for exponents in base_base:
            if not factor_added:
                corrected_factor = pow(self.factor, 1 / (exponents[1]))
                converted = [exponents[0], exponents[1], exponents[2] * corrected_factor]
                factor_added = True
            else:
                converted = exponents
            result.append(converted)
        return result


class PrefixedUnit(Unit):

    def __init__(self, prefix, base_unit, identifier=None, system_of_units=None, is_base_unit=False):
        label = f'{prefix.name}{base_unit.label()}'
        symbol = f'{prefix.symbol}{base_unit.symbol()}'
        self.prefix = prefix
        self.baseUnit = base_unit
        if system_of_units is None and not is_base_unit and base_unit is not None:
            system_of_units = base_unit.systemOfUnits
        super().__init__(label, symbol, dimensions=self.baseUnit.dimensions, identifier=identifier,
                         system_of_units=system_of_units, is_base_unit=is_base_unit)

    def __eq__(self, other):
        if isinstance(other, PrefixedUnit):
            if not isinstance(self.identifier, URIRef) or not isinstance(other.identifier, URIRef):
                if other.baseUnit == self.baseUnit and other.prefix == self.prefix:
                    return True
            else:
                return str(self.identifier) == str(other.identifier)
        return False

    def get_units_exponents(self):
        return [[self, 1, 1.0]]

    def get_base_units_exponents(self):
        base_base = self.baseUnit.get_base_units_exponents()
        result = []
        factor_added = False
        for exponents in base_base:
            if not factor_added:
                corrected_factor = pow(self.prefix.factor, 1 / (exponents[1]))
                converted = [exponents[0], exponents[1], exponents[2] * corrected_factor]
                factor_added = True
            else:
                converted = exponents
            result.append(converted)
        return result


class UnitMultiple(SingularUnit):

    def __init__(self, base_unit, factor=1.0, identifier=None, label=None, symbol=None, system_of_units=None,
                 is_base_unit=False):
        if label is None:
            label = f'{factor} {base_unit.label()}'
        if symbol is None:
            symbol = f'{factor}{base_unit.symbol()}'
        super().__init__(label, symbol, base_unit.dimensions, base_unit, factor, identifier,
                         system_of_units=system_of_units, is_base_unit=is_base_unit)


class CompoundUnit(Unit):

    def __init__(self, symbol, dimensions, identifier=None, system_of_units=None, is_base_unit=False):
        super().__init__(None, symbol, dimensions, identifier, system_of_units=system_of_units,
                         is_base_unit=is_base_unit)


class UnitMultiplication(CompoundUnit):

    def __init__(self, multiplier, multiplicand, identifier=None, system_of_units=None,
                 is_base_unit=False):
        dimensions = multiplier.dimensions * multiplicand.dimensions
        multiplier_str = str(multiplier.symbol())
        if isinstance(multiplier, CompoundUnit) and not isinstance(multiplier, UnitExponentiation):
            multiplier_str = f'{multiplier_str}'
        multiplicand_str = str(multiplicand.symbol())
        if isinstance(multiplicand, CompoundUnit) and not isinstance(multiplicand, UnitExponentiation):
            multiplicand_str = f'{multiplicand_str}'
        if system_of_units is None and multiplier.systemOfUnits is not None and multiplicand.systemOfUnits is not None \
                and multiplier.systemOfUnits == multiplicand.systemOfUnits:
            system_of_units = multiplier.systemOfUnits
        symbol = f'{multiplier_str}·{multiplicand_str}'
        super().__init__(symbol, dimensions, identifier, system_of_units=system_of_units, is_base_unit=is_base_unit)
        self.multiplier = multiplier
        self.multiplicand = multiplicand

    def __eq__(self, other):
        if isinstance(other, UnitMultiplication):
            if not isinstance(self.identifier, URIRef) or not isinstance(other.identifier, URIRef):
                if other.multiplier == self.multiplier and other.multiplicand == self.multiplicand:
                    return True
                if other.multiplier == self.multiplicand and other.multiplicand == self.multiplier:
                    return True
            else:
                return str(self.identifier) == str(other.identifier)
        return False

    def get_units_exponents(self):
        multiplier_exp = self.multiplier.get_units_exponents()
        multiplicand_exp = self.multiplicand.get_units_exponents()
        result = []
        result.extend(multiplier_exp)
        result.extend(multiplicand_exp)
        return result

    def get_base_units_exponents(self):
        multiplier_base = self.multiplier.get_base_units_exponents()
        multiplicand_base = self.multiplicand.get_base_units_exponents()
        result = []
        result.extend(multiplier_base)
        result.extend(multiplicand_base)
        return result


class UnitDivision(CompoundUnit):

    def __init__(self, numerator, denominator, identifier=None, system_of_units=None, is_base_unit=False):
        dimensions = numerator.dimensions / denominator.dimensions
        numerator_str = str(numerator.symbol())
        if isinstance(numerator, CompoundUnit) and not isinstance(numerator, UnitExponentiation):
            numerator_str = f'({numerator_str})'
        denominator_str = str(denominator.symbol())
        if isinstance(denominator, CompoundUnit) and not isinstance(denominator, UnitExponentiation):
            denominator_str = f'({denominator_str})'
        if system_of_units is None and numerator.systemOfUnits is not None and denominator.systemOfUnits is not None \
                and numerator.systemOfUnits == denominator.systemOfUnits:
            system_of_units = numerator.systemOfUnits
        symbol = f'{numerator_str}/{denominator_str}'
        super().__init__(symbol, dimensions, identifier, system_of_units=system_of_units, is_base_unit=is_base_unit)
        self.numerator = numerator
        self.denominator = denominator

    def __eq__(self, other):
        if isinstance(other, UnitDivision):
            if not isinstance(self.identifier, URIRef) or not isinstance(other.identifier, URIRef):
                if other.numerator == self.numerator and other.denominator == self.denominator:
                    return True
            else:
                return str(self.identifier) == str(other.identifier)
        return False

    def get_units_exponents(self):
        numerator_exp = self.numerator.get_units_exponents()
        denominator_exp = self.denominator.get_units_exponents()
        result = []
        result.extend(numerator_exp)
        for exponents in denominator_exp:
            result.append([exponents[0], -exponents[1], exponents[2]])
        return result

    def get_base_units_exponents(self):
        numerator_base = self.numerator.get_base_units_exponents()
        denominator_base = self.denominator.get_base_units_exponents()
        result = []
        result.extend(numerator_base)
        for exponents in denominator_base:
            result.append([exponents[0], -exponents[1], exponents[2]])
        return result


class UnitExponentiation(CompoundUnit):

    def __init__(self, base, exponent, identifier=None, system_of_units=None, is_base_unit=False):
        dimensions = Dimension.pow(base.dimensions, exponent)
        base_str = str(base.symbol())
        if isinstance(base, CompoundUnit):
            base_str = f'({base_str})'
        if system_of_units is None and base.systemOfUnits is not None:
            system_of_units = base.systemOfUnits
        symbol = f'{base_str}{exponent}'
        super().__init__(symbol, dimensions, identifier, system_of_units=system_of_units, is_base_unit=is_base_unit)
        self.base = base
        self.exponent = exponent

    def __eq__(self, other):
        if isinstance(other, UnitExponentiation):
            if not isinstance(self.identifier, URIRef) or not isinstance(other.identifier, URIRef):
                if other.base == self.base and other.exponent == self.exponent:
                    return True
            else:
                return str(self.identifier) == str(other.identifier)
        return False

    def get_units_exponents(self):
        base_exp = self.base.get_units_exponents()
        result = []
        for exponents in base_exp:
            result.append([exponents[0], exponents[1] * self.exponent, exponents[2]])
        return result

    def get_base_units_exponents(self):
        base_base = self.base.get_base_units_exponents()
        result = []
        for exponents in base_base:
            result.append([exponents[0], exponents[1] * self.exponent, exponents[2]])
        return result



import time
import unittest

import jwt
import requests_mock

from clients.keycloak_client import KeycloakClient


class AuthServiceClientUnitTest(unittest.TestCase):

    def response(self, username) -> dict:
        return dict({
            "client_id": username,
            "access_token": "eyEY1234"
        })

    def token(self, username: str, roles: [str], iat: int = int(time.time())) -> str:
        claims = {
            'iat': iat,
            'client_id': username,
            'realm_access': {
                'roles': roles
            }
        }
        with open('tests/rsa/rs256.key', 'rb') as fh:
            return jwt.JWT().encode(claims, jwt.jwk_from_pem(fh.read()), alg='RS256')

    def test_obtain_user_token_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('http://auth-service:8080/realms/dbrepo/protocol/openid-connect/token',
                      json=self.response("username"))
            # test
            token = KeycloakClient().obtain_user_token("username", "password")
            self.assertEqual("eyEY1234", token)

    def test_obtain_user_token_malformed_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('http://auth-service:8080/realms/dbrepo/protocol/openid-connect/token',
                      json={"client_id": "username"})
            # test
            try:
                KeycloakClient().obtain_user_token("username", "password")
                self.fail()
            except AssertionError:
                pass

    def test_verify_jwt_succeeds(self):
        # test
        user = KeycloakClient().verify_jwt(self.token("username", []))
        self.assertEqual("username", user.username)



import os
import unittest

import opensearchpy
from dbrepo.api.dto import Database, Table, Column, ColumnType, Constraints, PrimaryKey, \
    ConceptBrief, UnitBrief, UserBrief, ContainerBrief, ImageBrief, TableBrief, ColumnBrief
from opensearchpy import NotFoundError

from clients.opensearch_client import OpenSearchClient

req = Database(id="209acf92-5c9b-4633-ad99-113c86f6e948",
               name="Test",
               internal_name="test_tuw1",
               owner=UserBrief(id="c6b71ef5-2d2f-48b2-9d79-b8f23a3a0502", username="foo"),
               contact=UserBrief(id="c6b71ef5-2d2f-48b2-9d79-b8f23a3a0502", username="foo"),
               exchange_name="dbrepo",
               is_public=True,
               is_schema_public=True,
               is_dashboard_enabled=True,
               container=ContainerBrief(id="7efe8b27-6cdc-4387-80e3-92ee28f4a7c5",
                                        name="MariaDB",
                                        internal_name="mariadb",
                                        image=ImageBrief(id="f97791b4-baf4-4b18-8f7d-3084818e6549",
                                                         name="mariadb",
                                                         version="11.1.3",
                                                         default=True)),
               tables=[Table(id="f94a6164-cad4-4873-a9fd-3fe5313b2e95",
                             database_id="209acf92-5c9b-4633-ad99-113c86f6e948",
                             name="Data",
                             internal_name="data",
                             owner=UserBrief(id="c6b71ef5-2d2f-48b2-9d79-b8f23a3a0502", username="foo"),
                             constraints=Constraints(uniques=[], foreign_keys=[], checks=[], primary_key=[]),
                             is_versioned=False,
                             queue_name="dbrepo",
                             routing_key="dbrepo.1.1",
                             is_public=True,
                             is_schema_public=True,
                             columns=[Column(id="7bef7e68-88f1-438e-9b94-0a77afd21471",
                                             database_id="209acf92-5c9b-4633-ad99-113c86f6e948",
                                             table_id="f94a6164-cad4-4873-a9fd-3fe5313b2e95",
                                             name="ID",
                                             ord=0,
                                             internal_name="id",
                                             type=ColumnType.BIGINT,
                                             is_null_allowed=False,
                                             size=20,
                                             d=0,
                                             concept=ConceptBrief(id="fb32ecf6-1f68-49b4-85ee-04e76263cbef",
                                                                  uri="http://www.wikidata.org/entity/Q2221906"),
                                             unit=UnitBrief(id="a67d735e-32ef-4917-b412-fe099c6757a1",
                                                            uri="http://www.ontology-of-units-of-measure.org/resource/om-2/degreeCelsius"),
                                             val_min=0,
                                             val_max=10)]
                             )])


class SearchServiceClientUnitTest(unittest.TestCase):

    def test_update_database_succeeds(self):
        req.tables = [Table(id="f94a6164-cad4-4873-a9fd-3fe5313b2e95",
                            name="Test Table",
                            internal_name="test_table",
                            queue_name="dbrepo",
                            routing_key="dbrepo.test_tuw1.test_table",
                            is_public=True,
                            is_schema_public=True,
                            database_id=req.id,
                            constraints=Constraints(uniques=[],
                                                    foreign_keys=[],
                                                    checks=[],
                                                    primary_key=[PrimaryKey(id="f0d4dfdf-d987-4c73-aa40-1038db79bb31",
                                                                            table=TableBrief(
                                                                                id="f94a6164-cad4-4873-a9fd-3fe5313b2e95",
                                                                                database_id=req.id,
                                                                                name="Test Table",
                                                                                internal_name="test_table",
                                                                                is_public=True,
                                                                                is_schema_public=True,
                                                                                is_versioned=True,
                                                                                owned_by="c6b71ef5-2d2f-48b2-9d79-b8f23a3a0502"),
                                                                            column=ColumnBrief(
                                                                                id="90d81c6a-e69a-4413-97b8-fd2266a6e641",
                                                                                name="ID",
                                                                                database_id=req.id,
                                                                                table_id="90d81c6a-e69a-4413-97b8-fd2266a6e641",
                                                                                internal_name="id",
                                                                                type=ColumnType.BIGINT))]),
                            is_versioned=True,
                            owner=UserBrief(id="c6b71ef5-2d2f-48b2-9d79-b8f23a3a0502", username="foo"),
                            columns=[Column(id="c63bde4a-61e4-42f1-ab64-350579c0691f",
                                            database_id=req.id,
                                            table_id="f94a6164-cad4-4873-a9fd-3fe5313b2e95",
                                            ord=0,
                                            name="ID",
                                            internal_name="id",
                                            type=ColumnType.BIGINT,
                                            is_null_allowed=False)])]
        # mock
        OpenSearchClient().update_database(database_id=req.id, data=req)

        # test
        database = OpenSearchClient().update_database(database_id=req.id, data=req)
        self.assertEqual("209acf92-5c9b-4633-ad99-113c86f6e948", database.id)
        self.assertEqual("Test", database.name)
        self.assertEqual("test_tuw1", database.internal_name)
        self.assertEqual("c6b71ef5-2d2f-48b2-9d79-b8f23a3a0502", database.owner.id)
        self.assertEqual("foo", database.owner.username)
        self.assertEqual("c6b71ef5-2d2f-48b2-9d79-b8f23a3a0502", database.contact.id)
        self.assertEqual("foo", database.contact.username)
        self.assertEqual("dbrepo", database.exchange_name)
        self.assertEqual(True, database.is_public)
        self.assertEqual("7efe8b27-6cdc-4387-80e3-92ee28f4a7c5", database.container.id)
        # ...
        self.assertEqual("f97791b4-baf4-4b18-8f7d-3084818e6549", database.container.image.id)
        # ...
        self.assertEqual(1, len(database.tables))
        self.assertEqual("f94a6164-cad4-4873-a9fd-3fe5313b2e95", database.tables[0].id)
        self.assertEqual("Test Table", database.tables[0].name)
        self.assertEqual("test_table", database.tables[0].internal_name)
        self.assertEqual("dbrepo", database.tables[0].queue_name)
        self.assertEqual("dbrepo.test_tuw1.test_table", database.tables[0].routing_key)
        self.assertEqual(True, database.tables[0].is_public)
        self.assertEqual("209acf92-5c9b-4633-ad99-113c86f6e948", database.tables[0].database_id)
        self.assertEqual(True, database.tables[0].is_versioned)
        self.assertEqual("c6b71ef5-2d2f-48b2-9d79-b8f23a3a0502", database.tables[0].owner.id)
        self.assertEqual("foo", database.tables[0].owner.username)
        self.assertEqual(1, len(database.tables[0].columns))
        self.assertEqual("c63bde4a-61e4-42f1-ab64-350579c0691f", database.tables[0].columns[0].id)
        self.assertEqual("ID", database.tables[0].columns[0].name)
        self.assertEqual("id", database.tables[0].columns[0].internal_name)
        self.assertEqual(ColumnType.BIGINT, database.tables[0].columns[0].type)
        self.assertEqual("209acf92-5c9b-4633-ad99-113c86f6e948", database.tables[0].columns[0].database_id)
        self.assertEqual("f94a6164-cad4-4873-a9fd-3fe5313b2e95", database.tables[0].columns[0].table_id)
        self.assertEqual(False, database.tables[0].columns[0].is_null_allowed)

    def test_update_database_create_succeeds(self):
        # test
        database = OpenSearchClient().update_database(database_id=req.id, data=req)
        self.assertEqual("209acf92-5c9b-4633-ad99-113c86f6e948", database.id)
        self.assertEqual("Test", database.name)
        self.assertEqual("test_tuw1", database.internal_name)
        self.assertEqual("c6b71ef5-2d2f-48b2-9d79-b8f23a3a0502", database.owner.id)
        self.assertEqual("foo", database.owner.username)
        self.assertEqual("c6b71ef5-2d2f-48b2-9d79-b8f23a3a0502", database.contact.id)
        self.assertEqual("foo", database.contact.username)
        self.assertEqual("dbrepo", database.exchange_name)
        self.assertEqual(True, database.is_public)
        self.assertEqual("7efe8b27-6cdc-4387-80e3-92ee28f4a7c5", database.container.id)
        # ...
        self.assertEqual("f97791b4-baf4-4b18-8f7d-3084818e6549", database.container.image.id)
        # ...
        self.assertEqual(1, len(database.tables))

    def test_update_database_malformed_fails(self):
        os.environ['OPENSEARCH_USERNAME'] = 'i_do_not_exist'

        # test
        try:
            database = OpenSearchClient().update_database(database_id=req.id, data=req)
        except opensearchpy.exceptions.TransportError:
            pass

    def test_delete_database_fails(self):

        # test
        try:
            OpenSearchClient().delete_database(database_id="deadbeef-a5aa-49bb-87e7-6c6271731a1a")
        except opensearchpy.exceptions.NotFoundError:
            pass

    def test_delete_database_succeeds(self):
        # mock
        OpenSearchClient().update_database(database_id=req.id, data=req)

        # test
        OpenSearchClient().delete_database(database_id=req.id)

    def test_get_fields_for_index_database_succeeds(self):
        # mock
        OpenSearchClient().update_database(database_id=req.id, data=req)

        # test
        response = OpenSearchClient().get_fields_for_index(field_type="database")
        self.assertTrue(len(response) > 0)

    def test_get_fields_for_index_user_succeeds(self):
        # mock
        OpenSearchClient().update_database(database_id=req.id, data=req)

        # test
        response = OpenSearchClient().get_fields_for_index(field_type="user")
        self.assertTrue(len(response) > 0)

    def test_fuzzy_search_succeeds(self):
        # mock
        OpenSearchClient().update_database(database_id=req.id, data=req)

        # test
        OpenSearchClient().fuzzy_search(search_term="test_tuw")

    def test_unit_independent_search_fails(self):
        # mock
        OpenSearchClient().update_database(database_id=req.id, data=req)

        # test
        try:
            OpenSearchClient().unit_independent_search(0, 100, {
                "unit.uri": "http://www.ontology-of-units-of-measure.org/resource/om-2/degreeCelsius"})
            self.fail()
        except NotFoundError:
            pass
